/****************************************************************************
** COPYRIGHT (C):    1993 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          ChiLib
** FILE:             chiprioq.cpp
** PURPOSE:          Priority queue template
** VERSION           1.0
** LANGUAGE:         Borland C++ 3.1
** TARGET:           IBM PC
** PROGRAMMER:       Cay Horstmann (CSH)
** START DATE:       7-30-94 (CSH)
** UPDATE HISTORY:
**

8/22/94 CSH moved length() to class definition--BC 3.1 messes up line
numbers with inline member function templates

**
****************************************************************************/

/* IMPORT ******************************************************************/

#ifndef EXPORT
#include "chisetup.h"
#endif

#include "chiprioq.h"

EXPORT
#include "chiarray.h"

/* TYPES *******************************************************************/

#ifndef EXPORT

template<class X>
class Chi_PriorityQueue
/* PURPOSE:    a container from which elements are retrieved in increasing order
*/
{
public:

   Chi_PriorityQueue(int (*comp)(const X&, const X&), int size = 0);
   void insert(const X& x);
   X remove();
   int length() const { return _length; }

   void empty();
private:

   enum { DEFAULT_SIZE = 100 };

   void upheap(int);
   void downheap(int);

   Chi_Array<X> _heap;
   int (*_comp)(const X&, const X&);
   int _length;
};

//***************************************************************************

class Chi_PtrPriorityQueueBase
/* PURPOSE:    base class for pointer priority queue
*/
{
public:

   Chi_PtrPriorityQueueBase(int (*comp)(const void*, const void*), int size = 0);
   int length() const { return _length; }

   void empty();

protected:
   void insert(void* x);
   void* remove();

private:

   enum { DEFAULT_SIZE = 100 };

   void upheap(int);
   void downheap(int);

   Chi_PtrArray<void> _heap;
   int (*_comp)(const void*, const void*);
   int _length;
};

//............................................................................

template<class X>
class Chi_PtrPriorityQueue : public Chi_PtrPriorityQueueBase
/* PURPOSE:    a container from which elements are retrieved in increasing order
*/
{
public:

   Chi_PtrPriorityQueue(int (*comp)(const X*, const X*), int size = 0)
      : Chi_PtrPriorityQueueBase((int(*)(const void*, const void*))comp, size) {}
   void insert(X* x) { Chi_PtrPriorityQueueBase::insert(x); }
   X* remove() { return (X*) Chi_PtrPriorityQueueBase::remove(); }
};

//***************************************************************************

template<class X>
Chi_PriorityQueue<X>::Chi_PriorityQueue(int (*comp)(const X&, const X&), int size)
/* RECEIVES:   comp - comparison function for stored objects
               size - desired size of the underlying heap array
*/
:  _comp(comp), _length(0),
   _heap(1, size == 0 ? DEFAULT_SIZE : size)
{}

//............................................................................

template<class X>
void Chi_PriorityQueue<X>::insert(const X& x)
/* PURPOSE:    add a new value
   RECEIVES:   x - the value to insert
*/
{  _length++;
   _heap.set(_length, x);
   upheap(_length);
}

//............................................................................

template<class X>
X Chi_PriorityQueue<X>::remove()
/* PURPOSE:    remove the smallest value from the queue
   RETURNS:    the removed value
*/
{  if (_length > 0)
   {  X r = _heap[1];
      _heap[1] = _heap[_length];
      _length--;
      downheap(1);
      return r;
   }
   else
   {  static X r;
      return r;
   }
}

//............................................................................

template<class X>
void Chi_PriorityQueue<X>::empty()
/* PURPOSE:    remove all elements from the container
*/
{  _heap.empty();
   _length = 0;
}

//............................................................................

template<class X>
void Chi_PriorityQueue<X>::upheap(int k)
/* PURPOSE:    move a small element to its correct position towards the top of the heap
   RECEIVES:   k - the offset of the element to move
*/
{  if (k <= 1) return;
   X v = _heap[k];
   while (k > 1 && (*_comp)(_heap[k / 2], v) > 0)
   {  _heap[k] = _heap[k / 2];
      k = k / 2;
   }
   _heap[k] = v;
}

//............................................................................

template<class X>
void Chi_PriorityQueue<X>::downheap(int k)
/* PURPOSE:    move a large element to its correct position towards the bottom of the heap
   RECEIVES:   k - the offset of the element to move
*/
{  if (k > _length / 2) return;
   X v = _heap[k];
   Bool done = FALSE;
   while (k <= _length / 2 && !done)
   {  int j = 2 * k;
      if (j < _length && (*_comp)(_heap[j], _heap[j + 1]) > 0)
         j++;
      if ((*_comp)(_heap[j], v) > 0)
         done = TRUE;
      else
      {  _heap[k] = _heap[j];
         k = j;
      }
   }
   _heap[k] = v;
}

#endif

//***************************************************************************

Chi_PtrPriorityQueueBase::Chi_PtrPriorityQueueBase(int (*comp)(const void*, const void*), int size)
/* RECEIVES:   comp - comparison function for stored objects
               size - desired size of the underlying heap array
*/
:  _comp(comp), _length(0),
   _heap(1, size == 0 ? DEFAULT_SIZE : size)
{}

//............................................................................


void Chi_PtrPriorityQueueBase::insert(void* x)
/* PURPOSE:    add a new value
   RECEIVES:   x - the value to insert
*/
{  _length++;
   _heap.set(_length, x);
   upheap(_length);
}

//............................................................................


void* Chi_PtrPriorityQueueBase::remove()
/* PURPOSE:    remove the smallest value from the queue
   RETURNS:    the removed value
*/
{  if (_length > 0)
   {  void* r = _heap[1];
      _heap[1] = _heap[_length];
      _length--;
      downheap(1);
      return r;
   }
   else
      return 0;
}

//............................................................................


void Chi_PtrPriorityQueueBase::empty()
/* PURPOSE:    remove all elements from the container
*/
{  _heap.empty();
   _length = 0;
}

//............................................................................


void Chi_PtrPriorityQueueBase::upheap(int k)
/* PURPOSE:    move a small element to its correct position towards the top of the heap
   RECEIVES:   k - the offset of the element to move
*/
{  if (k <= 1) return;
   void* v = _heap[k];
   while (k > 1 && (*_comp)(_heap[k / 2], v) > 0)
   {  _heap[k] = _heap[k / 2];
      k = k / 2;
   }
   _heap[k] = v;
}

//............................................................................


void Chi_PtrPriorityQueueBase::downheap(int k)
/* PURPOSE:    move a large element to its correct position towards the bottom of the heap
   RECEIVES:   k - the offset of the element to move
*/
{  if (k > _length / 2) return;
   void* v = _heap[k];
   Bool done = FALSE;
   while (k <= _length / 2 && !done)
   {  int j = 2 * k;
      if (j < _length && (*_comp)(_heap[j], _heap[j + 1]) > 0)
         j++;
      if ((*_comp)(_heap[j], v) > 0)
         done = TRUE;
      else
      {  _heap[k] = _heap[j];
         k = j;
      }
   }
   _heap[k] = v;
}


/* TEST ********************************************************************/

#ifdef CHIPRIOQ_TEST

#include <stdlib.h>
#include <string.h>
#include "chistrng.h"

int main()
{  
#ifndef CHI_BC30
   Chi_PriorityQueue<Chi_String> q(Chi_String::compare);
#else
   int (*c)(const Chi_String&, const Chi_String&) = Chi_String::compare;
   Chi_PriorityQueue<Chi_String> q(c);
#endif
   const int TRY = 200;
   Chi_String s;
   int i;
   for (i = 1; i <= TRY; i++) 
   {  s.from_double(rand());
      q.insert(s);
   }
   for (i = 1; i <= TRY + 1; i++) cout << q.remove() << endl;

   Chi_PtrPriorityQueue<char> pq(strcmp);
   for (i = 1; i <= TRY; i++) 
   {  s.from_double(rand());
      pq.insert(s.new_c_array());
   }
   for (i = 1; i <= TRY + 1; i++) cout << pq.remove() << endl;


   return 0;
}

#endif

/***************************************************************************/
