/****************************************************************************
** COPYRIGHT (C):    1992...1994 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          ChiLib
** FILE:             chistrng.cpp
** PURPOSE:          a simple "value semantics" string class
** VERSION           1.3
** PROGRAMMER:       Cay Horstmann (CSH)
** START DATE:       6-10-92 (CSH)
** UPDATE HISTORY:

4-15-93 CSH
Cleaned up and added functions receiving const char[]

6-17-93 CSH
Changed logic to remove one level of indirection

8-23-93 CSH
Added field functions, read_line, operator const char*

9-22-93 CSH remove_field
Now returns the field without trailing spaces to exactly match the return
value of field(). (Trailing spaces are still removed.)

10-7-93 CSH set, []
Added call to unique to disassociate from shared strings before modification

1-11-94 CSH
Changed names to conform to proposed ANSI standard
Changed (from, to) range specs to (from, n) to conform to proposed ANSI
standard

4-28-94 CSH
Switched to chierror for error handling

10-5-95 CSH
Removed non-portable functions gcvt, ltoa and stricmp

****************************************************************************/

/* NOTES ********************************************************************

NOTE: C comments inside this comment are written |* like this *|

This class implements a simple string class with reference counts and delayed
copy. It is intended to be a complete replacement of C style char* pointers.

The name Chi_String for the class may be ugly, but every Tom, Dick and Harry
has a String class, and we don't want a name conflict.

The biggest difference between Chi_String objects and char* pointers is that
our strings manage their own storage, and that they have value semantics.
For example,

   Chi_String a("Harry");
   Chi_String b = a;
   b[1] = 'u';
   // now a is "Harry", b is "Hurry"

Contrast this with char* pointers.

   char* a = "Harry";
   char* b = a;
   b[1] = 'u';
   // now both a and b are "Hurry"


You should never have to think about storage management. This should come as
a welcome relief to the C programmer: no more

   char* s; // oops--no memory allocated
   strcpy(s, "Hello"); // Oh no!!!
   char t[4]; // not necessarily an empty string
   strcat(t, "World"); // Not again!!!

Chi_String objects allocate as much memory as needed, and they are initialized
to an empty string unless another string is explicitly given.

   Chi_String s;
   s = "Hello";
   Chi_String t;
   t.append("World");

Concatenation works with append or the + operator:

   Chi_String s = 'H';
   Chi_String t = "ell";
   Chi_String u = s + t;
   u.append('o');

Comparison operators work correctly, using lexicographic comparison.

   if ("Hello" <= s && s < "World"  || s == "") ...

Note that at least one of the arguments must be a Chi_String, the other
can be a regular string (but not a character).

Although quoted characters ('x') and strings ("xyz") are converted to
Chi_String objects through constructors, we supply several versions of
member functions with String, char and const char[] arguments to bypass
the construction and immediate destruction of a temporary for added
efficiency. This makes the list of member functions quite long.

Occasionally, you may need to use prehistoric functions that require char*
pointers rather than Chi_Strings, for example to call a library routine like
fopen.

There are three ways of getting a char* out of a Chi_String.

   Chi_String s;

   // method 1 -- preferred
   char buffer[80];
   s.c_array(buffer, sizeof(buffer));
   fp = fopen(buffer, "r");

   // method 2 -- get pointer to internal representation
   // This pointer is VERY short lived. NEVER store the result.
   fopen(s.c_str(), "r"); // OK
   const char* p = s.c_str(); // NO!

   // method 3 -- remember to delete
   char* p = s.new_c_array();
   fopen(p, "r");
   delete[] p;

The second method is the simplest, but it is inherently dangerous. The
type conversion returns a const char* pointer into the innards of a Chi_String.
If another function messes up this memory, weird things will happen. Also, the
lifetime of this memory is unpredictable. ANY string operation can cause
relocation of the memory block.

Some extremely prehistoric library functions (like TextOut in the 3.0 version
of <windows.h>) have prototypes of char* for all strings, even those they do
not intend to modify. In that case, write (|* non-const *| char*) s.c_str()
and place a curse on the library vendor.

If you must use Chi_String::c_str(), follow these three rules.

   (1) NEVER store the pointer returned by c_str anywhere. Its lifetime is
       short and unpredictable. NEVER, NEVER return it as a function value
   (2) NEVER write anything to that pointer. You may mess up more than one
       string because of the delayed copy mechanism
   (3) NEVER call c_str together with another string function in the same
       statement

The following table shows the Chi_String equivalents of all ANSI string
functions

string.h                    str.h
---------------------------------------------
strcpy(s, t)                s = t
strncpy(s, t, n)            s = t.substr(0, n)
strcat(s, t)                s.append(t) or s += t
strcmp(s, t)                s.compare(t) or s relop t
                              (relop can be ==, !=, <, <=, >, >=)
strncmp(s, t, n)            s.compare(t, n)
                              or s.substr(0, n) relop t.substr(0, n)
strchr(s, c)                s.find(c)
strrchr(s, c)               s.rfind(c)
strspn(s, t)                s.find_first_not_of(t)
strcspn(s, t)               s.find_first_of(t)
strpbrk(s, t)               s.find_first_of(t)
strstr(s, t)                s.find(t)
strtok(s, t1), then         i = s.find_first_of(t), then
strtok(0, t2)               s.find_first_of(t2, i)

TECHNICAL NOTES:

Returning pointers into an array that may be reallocated is always
problematic. The string class has two such functions: operator[] and c_str().
Neither of these functions should be used in a context in which reallocation
may occur. For this reason, operator[] does not grow the array. If it did,
statements like a[i] = f(a[i], a[i+10]) would fail mysteriously if the
compiler evaluated the reference a[i], and then a[i+10], and the array was
relocated during the second computation. Similarly, for this reason, all
functions that relocate the array have a void return value to avoid nested
calls like f(a[i], a.insert(10, b).get(11)). DO NOT try to overcome
this with the comma operator.

****************************************************************************/

/* IMPORT ******************************************************************/

#ifndef EXPORT
#include "chisetup.h"
#endif

#include <stdlib.h>
EXPORT
#include <string.h>
#include <ctype.h>
#include <float.h>

EXPORT
#include <limits.h>

EXPORT
#include <iostream.h>

#include "chierror.h"

#include <stdio.h>
#include "chistrng.h"

/* TYPES *******************************************************************/

#ifndef EXPORT

class Chi_String
/* PURPOSE:       A string class
*/
{
public:
   enum {NPOS = INT_MAX}; // maximum number of positions in string

   Chi_String();                       // constructs an empty string
   Chi_String(const char s[]);
   Chi_String(char ch);

   char& operator[](int i);
   void set(int i, char ch);

   void insert(int at, const Chi_String& s);
   void insert(int at, const char s[]);
   void insert(int at, char ch);
   void insert_field(int, const Chi_String&);

   void append(const Chi_String& s);
   void append(const char s[]);
   void append(char ch);
   const Chi_String& operator+=(const Chi_String& s);
   const Chi_String& operator+=(const char s[]);
   const Chi_String& operator+=(char ch);

   Chi_String remove(int from, int n = NPOS);
   char remove_char(int at);
   Chi_String remove_field(int);

   void replace(int from, int n, const Chi_String& s);
   void replace(int from, int n, const char s[]);
   void replace(int from, int n, char ch);
   Chi_String replace_field(int, const Chi_String&);

   Chi_String& to_lower();
   Chi_String& to_upper();
   void from_long(long n);
   void from_double(double d);

   void grow(int n);
   void shrink(int n);
   void empty();

   int read_line(istream& is);
   void read_delimited(istream& is);

   int length() const;
   Bool is_empty() const;

   char operator[](int i) const;
   char get(int i) const;
   Chi_String substr(int from, int n = NPOS) const;
   Chi_String field(int) const;

   char* c_array(char buf[], size_t buflen, int from = 0, int n = NPOS) const;
   const char* c_str() const;
   char* new_c_array() const;

#ifdef CHI_SC70
   int compare_to(const Chi_String&) const;
   int compare_insens_to(const Chi_String&) const;
   int compare_to(const char s[]) const;
   int compare_to(const Chi_String&, int n) const;
   int compare_to(const char s[], int n) const;
   int compare_insens_to(const char s[]) const;
   // Symantec 7.0 can't handle name conflict with static function
#else
   int compare(const Chi_String&) const;
   int compare_insens(const Chi_String&) const;
   int compare(const char s[]) const;
   int compare(const Chi_String&, int n) const;
   int compare(const char s[], int n) const;
   int compare_insens(const char s[]) const;
#endif

   int find(char c, int from = 0) const;
   int find(const Chi_String& s, int from = 0) const;
   int find(const char s[], int from = 0) const;

   int rfind(char c, int from = NPOS) const;

   int find_first_of(const Chi_String& s, int from = 0) const;
   int find_first_of(const char s[], int from = 0) const;
   int find_first_not_of(const Chi_String& s, int from = 0) const;
   int find_first_not_of(const char s[], int from = 0) const;
   int find_last_of(const Chi_String& s, int from = 0) const;
   int find_last_of(const char s[], int from = 0) const;

   long to_long() const;
   double to_double() const;

   void write_delimited(ostream& os) const;

   Chi_String operator+(const Chi_String&) const;
   Chi_String operator+(const char b[]) const;
   Chi_String operator+(char) const;

   static int compare(const Chi_String& a, const Chi_String& b);
   static int compare_insens(const Chi_String& a, const Chi_String& b);
   static int hash(const Chi_String& a);

   friend Chi_String operator+(const char a[], const Chi_String& b);
   friend Chi_String operator+(char ch, const Chi_String& b);

   Chi_String(const Chi_String&);
   const Chi_String& operator=(const Chi_String&);
   ~Chi_String();

private:
   Chi_String(size_t len, const char s[], size_t slen);
   Chi_String(int);  /* left undefined--to avoid accidental conversion
                        int -> char -> Chi_String
                     */
   void detach();
   void unique();

   void field_range(int f, int& from, int& len) const;

   char* _str;
};

/*-------------------------------------------------------------------------*/

class Chi_StringHeader
/* PURPOSE:    The header block with reference count and available space
   REMARKS:    An instance of this structure is BELOW each string
               that is allocated on the collectible heap
*/
{
public:
   int unique() { return _rc == 1; }
   void incref() { ++_rc; }
   int decref() { return --_rc; }

   size_t length() const { return _length; }

   static char* alloc(const char[]);
   static char* alloc(int, const char[], int);
   static void free(char*);

   static Chi_StringHeader* header(const char* p) { return (Chi_StringHeader*)p - 1; }

private:
   char* start() const { return (char*)(this+1); }

   unsigned _rc;                       // reference count
   size_t _length;

friend class Chi_String;
};
#endif

/* FUNCTIONS ***************************************************************/

static int chi_stricmp(const char s[], const char t[])
/* PURPOSE:    lexicographically compares two strings without regard to case
   RECEIVES:   s, t - strings to compare
   RETURNS:    < 0 if s before t
               0 if s equals t, ignoring case
               > 0 if s after t
*/
{  int i = 0;
   while ( s[i] != 0 && t[i] != 0 && toupper(s[i]) == toupper(t[i])) 
   {  i++;
   }
   return toupper(s[i]) - toupper(t[i]);
}

/*.........................................................................*/

EXPORT Chi_String operator+(const char a[], const Chi_String& b)
/* PURPOSE:    concatenate two strings
   RECEIVES:   a, b - the strings to concatenate
   RETURNS:    the concatenated string
*/
{  int alen = a ? strlen(a) : 0;
   Chi_String r(alen + b.length(), a, alen);
   if (b._str) strcpy(r._str + alen, b._str);
   return r;
}

/*.........................................................................*/

EXPORT Chi_String operator+(char ch, const Chi_String& b)
/* PURPOSE:    concatenate a character and a string
   RECEIVES:   ch, b - the character and string to concatenate
   RETURNS:    the concatenated string
*/
{  Chi_String r(1 + b.length(), &ch, 1);
   if (b._str) strcpy(r._str + 1, b._str);
   return r;
}

/*.........................................................................*/

EXPORT ostream& operator<<(ostream& os, const Chi_String& s)
/* PURPOSE:    Print a string on a stream
   RECEIVES:   os - the stream
               s - the string to print
   RETURNS:    os
*/
{  return os << s.c_str();
}

/*.........................................................................*/

EXPORT istream& operator>>(istream& is, Chi_String& s)
/* PURPOSE:    Read a string from a stream
   RECEIVES:   is - the stream
               s (OUT) - the string that is read
   RETURNS:    is
*/
{  const int BUFLEN = 80;
   char buffer[ BUFLEN ];
   ws(is); // skip whitespace
   if (!is) return is;
   s.empty();
   do
   {  is.width(sizeof(buffer));
      is >> buffer;
      if (is) s.append(buffer);
      else is.width(0);
   }
   while (is && !isspace(is.peek()));
   return is;
}

/*-------------------------------------------------------------------------*/

#ifndef EXPORT

inline Chi_String::Chi_String()
/* REMARKS:    Constructs an empty string
*/
:  _str(0)
{}

/*...........................................................................*/

inline Chi_String::Chi_String(const char s[])
/* RECEIVES:   s - C character array
*/
:  _str(Chi_StringHeader::alloc(s))
{}

/*...........................................................................*/

inline Chi_String::Chi_String(char ch)
/* RECEIVES:   ch - single character
*/
:  _str(Chi_StringHeader::alloc(1, &ch, 1))
{}

/*...........................................................................*/

inline Chi_String::Chi_String(size_t tlen, const char s[], size_t slen)
/* RECEIVES:      tlen - total length of the string
                  s - character string to initialize first bytes
                  slen - length of s
   REMARKS:       (1) The string gets a 0 at position len
                  (2) This constructor is private, and its interface is made
                      slightly obscure on purpose to avoid accidental invocation
                      by class users. Originally we had a Chi_String(size_t),
                      and under some circumstances the compiler complained about
                      ambiguities with String(char)
*/
:  _str(Chi_StringHeader::alloc(tlen, s, slen))
{}

/*.........................................................................*/

inline Chi_String::Chi_String(const Chi_String& b)
:  _str(b._str)
{  if (_str) Chi_StringHeader::header(_str)->incref(); }

/*...........................................................................*/

inline void Chi_String::empty()
/* PURPOSE:    Empties the string
*/
{  detach();
   _str = 0;
}

/*.........................................................................*/

inline Chi_String::~Chi_String()
{  detach();
}

/*...........................................................................*/

inline int Chi_String::length() const
/* RETURNS:       the length of the string
*/
{  return _str ? Chi_StringHeader::header(_str)->_length : 0;
}

/*...........................................................................*/

inline Bool Chi_String::is_empty() const
/* RETURNS:       TRUE iff the string is empty
*/
{  return length() == 0;
}

/*...........................................................................*/

inline char Chi_String::get(int i) const
/* RETURNS:       the ith character in a string (0 = first character)
   REMARKS:       returns 0 if index is out of bounds
*/
{  return 0 <= i && i < length() ? _str[i] : 0;
}

//............................................................................

inline const char* Chi_String::c_str() const
/* PURPOSE:    get an internal pointer to the content of the string
   RETURNS:    the character array of the internal storage
   REMARKS:    the returned pointer is short-lived and will become
               invalid if self is modified or destroyed
*/
{  return _str ? _str : "";
}

//............................................................................

inline int operator==(const Chi_String& a, const Chi_String& b)
{ return strcmp(a.c_str(), b.c_str()) == 0; }
inline int operator!=(const Chi_String& a, const Chi_String& b)
{ return strcmp(a.c_str(), b.c_str()) != 0; }
inline int operator<(const Chi_String& a, const Chi_String& b)
{ return strcmp(a.c_str(), b.c_str()) < 0; }
inline int operator<=(const Chi_String& a, const Chi_String& b)
{ return strcmp(a.c_str(), b.c_str()) <= 0; }
inline int operator>(const Chi_String& a, const Chi_String& b)
{ return strcmp(a.c_str(), b.c_str()) > 0; }
inline int operator>=(const Chi_String& a, const Chi_String& b)
{ return strcmp(a.c_str(), b.c_str()) >= 0; }

inline int operator==(const Chi_String& a, const char b[])
{ return strcmp(a.c_str(), b ? b : "") == 0; }
inline int operator!=(const Chi_String& a, const char b[])
{ return strcmp(a.c_str(), b ? b : "") != 0; }
inline int operator<(const Chi_String& a, const char b[])
{ return strcmp(a.c_str(), b ? b : "") < 0; }
inline int operator<=(const Chi_String& a, const char b[])
{ return strcmp(a.c_str(), b ? b : "") <= 0; }
inline int operator>(const Chi_String& a, const char b[])
{ return strcmp(a.c_str(), b ? b : "") > 0; }
inline int operator>=(const Chi_String& a, const char b[])
{ return strcmp(a.c_str(), b ? b : "") >= 0; }

inline int operator==(const char a[], const Chi_String& b)
{ return strcmp(a ? a : "", b.c_str()) == 0; }
inline int operator!=(const char a[], const Chi_String& b)
{ return strcmp(a ? a : "", b.c_str()) != 0; }
inline int operator<(const char a[], const Chi_String& b)
{ return strcmp(a ? a : "", b.c_str()) > 0; }
inline int operator<=(const char a[], const Chi_String& b)
{ return strcmp(a ? a : "", b.c_str()) >= 0; }
inline int operator>(const char a[], const Chi_String& b)
{ return strcmp(a ? a : "", b.c_str()) < 0; }
inline int operator>=(const char a[], const Chi_String& b)
{ return strcmp(a ? a : "", b.c_str()) <= 0; }

#endif

/*...........................................................................*/

char Chi_String::operator[](int i) const
/* RETURNS:       the ith character in a string (0 = first character)
   REMARKS:       fails if i out of bounds
*/
{  CHI_ASSERT_BOUNDS(i, 0, length() - 1);
   return _str[i];
}

/*...........................................................................*/

char& Chi_String::operator[](int i)
/* RETURNS:       a reference to the ith character in a string (0 = first character)
   REMARKS:       fails if i out of bounds
                  It does NOT grow the string
*/
{  CHI_ASSERT_BOUNDS(i, 0, length() - 1);
   unique();
   return _str[i];
}

/*.........................................................................*/

const Chi_String& Chi_String::operator=(const Chi_String& b)
{  if (b._str) Chi_StringHeader::header(b._str)->incref();
   detach();
   _str = b._str;
   return self;
}

/*...........................................................................*/

void Chi_String::set(int i, char ch)
/* PURPOSE:       set a character in the string
   RECEIVES:      i - the position
                  ch - the new character
   REMARKS:       grows the string if necessary
*/
{  if (i >= 0)
   {  grow(i + 1);
      unique();
      _str[i] = ch;
   }
}

//............................................................................

void Chi_String::insert(int at, const Chi_String& s)
/* PURPOSE:    Insert a string inside self
   RECEIVES:   at - the position before which to insert
               s - the string to insert
*/
{  int len = length();
   if (at < 0) return;
   int slen = s.length();
   if (slen > 0)
   {  if (at <= len)
      {  Chi_String n(len + slen, _str, at);
         strncpy(n._str + at, s._str, slen);
         strncpy(n._str + at + slen, _str+at+1, len - at);
         self = n;
      }
      else
      {  Chi_String n(at + slen, _str, len);
         memset(n._str + len, ' ', at - len);
         strncpy(n._str + at, s._str, slen);
         self = n;
      }

   }
}

//............................................................................

void Chi_String::insert(int at, const char s[])
/* PURPOSE:    Insert a string inside self
   RECEIVES:   at - the position before which to insert
               s - the string to insert
*/
{  int len = length();
   if (at < 0 || s == 0) return;
   int slen = strlen(s);
   if (slen > 0)
   {  if (at <= len)
      {  Chi_String n(len + slen, _str, at);
         strncpy(n._str + at, s, slen);
         strncpy(n._str + at + slen, _str+at, len - at);
         self = n;
      }
      else
      {  Chi_String n(at + slen, _str, len);
         memset(n._str + len, ' ', at - len);
         strncpy(n._str + at, s, slen);
         self = n;
      }
   }
}

//............................................................................

void Chi_String::insert(int at, char ch)
/* PURPOSE:    Insert a character inside self
   RECEIVES:   at - the position before which to insert
               ch - the character to insert
*/
{  int len = length();
   if (at < 0) return;
   if (at <= len)
   {  Chi_String n(len + 1, _str, at);
      n._str[ at ] = ch;
      strncpy(n._str + at + 1, _str+at, len - at);
      self = n;
   }
   else
   {  Chi_String n(at + 1, _str, len);
      memset(n._str + len, ' ', at - len);
      n._str[ at ] = ch;
      self = n;
   }
}

/*...........................................................................*/

void Chi_String::append(const Chi_String& b)
/* PURPOSE:    append a string to self
   RECEIVES:   b - the string to append
*/
{  int len = length();
   Chi_String r(len + b.length(), _str, len);
   if (b._str) strcpy(r._str + length(), b._str);
   self = r;
}

/*...........................................................................*/

void Chi_String::append(const char b[])
/* PURPOSE:    append a string to self
   RECEIVES:   b - the string to append
*/
{  if (b == 0) return;
   int len = length();
   Chi_String r(len + (int)strlen(b), _str, len);
   if (r._str) strcpy(r._str + len, b);
   self = r;
}

/*...........................................................................*/

void Chi_String::append(char ch)
/* PURPOSE:    append a character to self
   RECEIVES:   ch - the character to append
   RETURNS:    self
   REMARKS:    this is slow since the entire string must be copied.
               Do not use this function to build up a string. Instead,
               use grow() to desired size and set() when possible
*/
{  int len = length();
   Chi_String r(len + 1, _str, len);
   r._str[ len ] = ch;
   self = r;
}

/*...........................................................................*/

#ifndef EXPORT

inline const Chi_String& Chi_String::operator+=(const Chi_String& b)
/* PURPOSE:    append a string to self
   RECEIVES:   b - the string to append
   RETURNS:    self
*/
{  append(b);
   return self;
}

/*...........................................................................*/

inline const Chi_String& Chi_String::operator+=(const char b[])
/* PURPOSE:    append a string to self
   RECEIVES:   b - the string to append
   RETURNS:    self
*/ 
{  append(b);
   return self;
}

/*...........................................................................*/

inline const Chi_String& Chi_String::operator+=(char ch)
/* PURPOSE:    append a character to self
   RECEIVES:   ch - the character to append
   RETURNS:    self
   REMARKS:    this is slow since the entire string must be copied.
               Do not use this function to build up a string. Instead,
               use grow() to desired size and set() when possible
*/
{  append(ch);
   return self;
}

#endif

//............................................................................

Chi_String Chi_String::remove(int from, int n)
/* PURPOSE:       Remove a substring from a string
   RECEIVES:      from - the starting index (0 = from beginning of the string)
                  n - the number of characters to remove
   RETURNS:       the removed string
   REMARKS:       if from + n >= string length, takes characters until end of string
*/
{  int len = length();
   if (from < 0) from = 0;
   if (n == NPOS || from + n >= len) n = len - from;
   Chi_String r(n, _str + from, n);
   if (n >= 0)
   {  Chi_String s(len - n, _str, from);
      strncpy(s._str + from, _str + from + n, len - from - n);
      self = s;
   }
   return r;
}

//............................................................................

char Chi_String::remove_char(int at)
/* PURPOSE:       Remove a character from a string
   RECEIVES:      at - the index (0 = beginning of the string)
   RETURNS:       the removed character
   REMARKS:       if at not in string, does nothing and returns 0
*/
{  int len = length();
   if (at < 0 || at >= len) return 0;
   char ch = _str[ at ]; 
   Chi_String n(len - 1, _str, at);
   strncpy(n._str + at, _str+at+1, len - at - 1);
   self = n;
   return ch;
}

//............................................................................

void Chi_String::replace(int from, int n, const Chi_String& s)
/* PURPOSE:       Replace a substring of self with another string
   RECEIVES:      from - the start of the substring to be replaced
                  n - the number of characters to be replaced
                  s - the string to be inserted
*/
{  int len = length();
   if (from < 0) from = 0;
   if (n == NPOS || from + n >= len) n = len - from;
   if (n > 0)
   {  int slen = s.length();
      Chi_String t(len - n + slen, _str, from);
      if (slen) strncpy(t._str + from, s._str, slen);
      strncpy(t._str + from + slen, _str + from + n, len - from - n);
      self = t;
   }
}

//............................................................................

void Chi_String::replace(int from, int n, const char s[])
/* PURPOSE:       Replace a substring of self with another string
   RECEIVES:      from - the start of the substring to be replaced
                  n - the number of characters to be replaced
                  s - the string to be inserted
*/
{  int len = length();
   if (from < 0) from = 0;
   if (n == NPOS || from + n >= len) n = len - from;
   if (n > 0)
   {  int slen = s ? strlen(s) : 0;
      Chi_String t(len - n + slen, _str, from);
      if (slen) strcpy(t._str + from, s);
      strncpy(t._str + from + slen, _str + from + n, len - from - n);
      self = t;
   }
}

//............................................................................

void Chi_String::replace(int from, int n, char ch)
/* PURPOSE:       Replace a substring of self with a single character
   RECEIVES:      from - the start of the substring to be replaced
                  n - the number of characters to be replaced
                  ch - the character to be inserted
*/
{  int len = length();
   if (from < 0) from = 0;
   if (n == NPOS || from + n >= len) n = len - from;
   if (n > 0)
   {  Chi_String t(len - n + 1, _str, from);
      t._str[ from ] = ch;
      strncpy(t._str + from + 1, _str + from + n, len - from - n);
      self = t;
   }
}

//............................................................................

Chi_String& Chi_String::to_lower()
/* PURPOSE:    Change all characters in self to lowercase
   RETURNS:    self
*/ 
{  unique();
   int len = length();
   for (int i = 0; i < len; i++)
      _str[i] = tolower(_str[i]);
   return self;
}

//............................................................................

Chi_String& Chi_String::to_upper()
/* PURPOSE:    Change all characters in self to uppercase
   RETURNS:    self
*/
{  unique();
   int len = length();
   for (int i = 0; i < len; i++)
      _str[i] = toupper(_str[i]);
   return self;
}

/*...........................................................................*/

void Chi_String::grow(int n)
/* PURPOSE:    Grows the string to have at least n characters
   RECEIVES:   n - minimum number of characters
   REMARKS:    if the current string has less than n characters, it is
               padded with spaces
*/
{  int len = length();
   if (len < n)
   {  Chi_String r(n, _str, len);
      memset(r._str + len, ' ', n - len);
      self = r;
   }
}

/*...........................................................................*/

void Chi_String::shrink(int n)
/* PURPOSE:    Shrinks the string to have at most n characters
   RECEIVES:   n - maximum number of characters
*/
{  if (n == 0) empty();
   else if (length() > n && n > 0)
   {  Chi_String r(n , _str, n);
      self = r;
   }
}

//............................................................................

char* Chi_String::c_array(char buf[], size_t buflen, int from, int n) const
/* PURPOSE:    get the content of the string with given range
   RECEIVES:   buf - the buffer to hold the string
               buflen - buffer size
               from - the char position to start copying
               n - the number of characters to copy
   RETURNS:    buf
   REMARKS:    if "from" and "to" is not provided, then get the whole string
*/
{  if (buf)
   {  int len = length();
      if (from < 0) from = 0;
      if (n == NPOS || from + n >= len) n = len - from;
      if (n >= 0)
      {  if (n >= buflen) n = buflen-1;
         strncpy(buf, _str, n);
         buf[n] = 0;
      }
      else buf[0] = 0;
   }
   return buf;
}

//............................................................................

char* Chi_String::new_c_array() const
/* PURPOSE: return a copy of the string on the heap
   REMARKS: since the function use "new" to allocate memory for buffer, user
            should "delete" it after he finished using the buffer.
*/
{  char* buffer = new char[length() + 1];
   if (buffer) 
   {  if (_str)
         strcpy(buffer, _str);
      else
         buffer[0] = 0;
   }
   return buffer;
}

//............................................................................

#ifdef CHI_SC70
int Chi_String::compare_to(const Chi_String& b) const
#else
int Chi_String::compare(const Chi_String& b) const
#endif
/* PURPOSE:    lexicographically compares two strings
   RECEIVES:   b - string to compare with self
   RETURNS:    < 0 if self before b
               0 if self equals b
               > 0 if self after b
*/  
{  return strcmp(c_str(), b.c_str());
}

//............................................................................

#ifdef CHI_SC70
int Chi_String::compare_to(const char s[]) const
#else
int Chi_String::compare(const char s[]) const
#endif
/* PURPOSE:    lexicographically compares two strings
   RECEIVES:   b - string to compare with self
   RETURNS:    < 0 if self before b
               0 if self equals b
               > 0 if self after b
*/  
{  return strcmp(c_str(), s ? s : "");
}

//............................................................................

#ifdef CHI_SC70
int Chi_String::compare_to(const Chi_String& b, int n) const
#else
int Chi_String::compare(const Chi_String& b, int n) const
#endif
/* PURPOSE:    lexicographically compares two strings
   RECEIVES:   b - string to compare with self
               n - max number of characters to compare
   RETURNS:    < 0 if self before b
               0 if self equals b
               > 0 if self after b
*/
{  return strncmp(c_str(), b.c_str(), n);
}

//............................................................................

#ifdef CHI_SC70
int Chi_String::compare_to(const char s[], int n) const
#else
int Chi_String::compare(const char s[], int n) const
#endif
/* PURPOSE:    lexicographically compares two strings
   RECEIVES:   b - string to compare with self
               n - max number of characters to compare
   RETURNS:    < 0 if self before b
               0 if self equals b
               > 0 if self after b
*/  
{  return strncmp(c_str(), s ? s : "", n);
}

//............................................................................


#ifdef CHI_SC70
int Chi_String::compare_insens_to(const Chi_String& b) const
#else
int Chi_String::compare_insens(const Chi_String& b) const
#endif
/* PURPOSE:    lexicographically compares two strings without regard to case
   RECEIVES:   b - string to compare with self
   RETURNS:    < 0 if self before b
               0 if self equals b, ignoring case
               > 0 if self after b
*/
{  return chi_stricmp(c_str(), b.c_str());
}

//............................................................................

#ifdef CHI_SC70
int Chi_String::compare_insens_to(const char s[]) const
#else
int Chi_String::compare_insens(const char s[]) const
#endif
/* PURPOSE:    lexicographically compares two strings without regard to case
   RECEIVES:   b - string to compare with self
   RETURNS:    < 0 if self before b
               0 if self equals b, ignoring case
               > 0 if self after b
*/
{  return chi_stricmp(c_str(), s ? s : "");
}

//............................................................................

int Chi_String::find(char c, int from) const
/* PURPOSE:    finds a character in a string
   RECEIVES:   c - the character to find
               from - the starting position (default = 0)
   RETURNS:    the index of the character
               NPOS if not found
*/
{  if (from < 0 ||  from >= length()) return NPOS;
   char* p = strchr(_str + from, c);
   return p == 0 ? NPOS : p - _str;
}

//............................................................................

int Chi_String::rfind(char c, int from) const
/* PURPOSE:    finds a character in a string, searching backwards
   RECEIVES:   c - the character to find
               from - the starting position (default = end of string)
   RETURNS:    the index of the character
               NPOS if not found
*/
{  int len = length();
   if (from >= len) from = len-1;
   while (from >= 0)
      if (_str[ from ] == c) return from;
      else from--;
   return NPOS;
}

//............................................................................

int Chi_String::find(const Chi_String& s, int from) const
/* PURPOSE:    finds a substring in a string
   RECEIVES:   s - the substring to find
               from - the starting position (default = 0)
   RETURNS:    the index of the start of the substring
               NPOS if not found
*/
{  if (from < 0 || from >= length()) return NPOS;
   char* p = strstr(_str + from, s.c_str());
   return p == 0 ? NPOS : p - _str;
}

//............................................................................

int Chi_String::find(const char s[], int from) const
/* PURPOSE:    finds a substring in a string
   RECEIVES:   s - the substring to find
               from - the starting position (default = 0)
   RETURNS:    the index of the start of the substring
               NPOS if not found
*/
{  if (from < 0 || from >= length()) return NPOS;
   char* p = strstr(_str + from, s ? s : "");
   return p == 0 ? NPOS : p - _str;
}

//............................................................................

int Chi_String::find_first_of(const Chi_String& s, int from) const
/* PURPOSE:    finds the first occurrence of a character from a set
   RECEIVES:   s - the set of characters to match
               from - the starting position (default = 0)
   RETURNS:    the index of the first character found
               NPOS if not found
*/
{  if (from < 0 || from >= length()) return NPOS;
   char* p = strpbrk(_str + from, s.c_str());
   return p == 0 ? NPOS : p - _str;
}

//............................................................................

int Chi_String::find_first_of(const char s[], int from) const
/* PURPOSE:    finds the first occurrence of a character from a set
   RECEIVES:   s - the set of characters to match
               from - the starting position (default = 0)
   RETURNS:    the index of the first character found
               NPOS if not found
*/
{  if (from < 0 || from >= length()) return NPOS;
   char* p = strpbrk(_str + from, s ? s : "");
   return p == 0 ? NPOS : p - _str;
}

//............................................................................

int Chi_String::find_first_not_of(const Chi_String& s, int from) const
/* PURPOSE:    finds the first occurrence of a character not in set
   RECEIVES:   s - the set of characters to avoid
               from - the starting position (default = 0)
   RETURNS:    the index of the first character found
               NPOS if not found
*/
{  if (from < 0 || from >= length()) return NPOS;
   size_t n = strspn(_str + from, s.c_str());
   return n == length() - from ? NPOS : from + n;
}

//............................................................................

int Chi_String::find_first_not_of(const char s[], int from) const
/* PURPOSE:    finds the first occurrence of a character not in set
   RECEIVES:   s - the set of characters to avoid
               from - the starting position (default = 0)
   RETURNS:    the index of the first character found
               NPOS if not found
*/
{  if (from < 0 || from >= length()) return NPOS;
   size_t n = strspn(_str + from, s ? s : "");
   return n == length() - from ? NPOS : from + n;
}

//............................................................................

int Chi_String::find_last_of(const Chi_String& s, int from) const
/* PURPOSE:    finds the last occurrence of a character from a set
   RECEIVES:   s - the set of characters to match
               from - the starting position (default = 0)
   RETURNS:    the index of the last character found
               NPOS if not found
*/
{  return find_last_of(s.c_str(), from);
}

//............................................................................

int Chi_String::find_last_of(const char s[], int from) const
/* PURPOSE:    finds the last occurrence of a character from a set
   RECEIVES:   s - the set of characters to match
               from - the starting position (default = 0)
   RETURNS:    the index of the last character found
               NPOS if not found
*/
{  if (from < 0 || from >= length()) return NPOS;
   char* p = _str + from;
   char* plast = 0;
   do
   {  p = strpbrk(p, s);
      if (p != 0) { plast = p; p++; }
   } while (p != 0);
   return plast == 0 ? NPOS : plast - _str;
}

/*...........................................................................*/

Chi_String Chi_String::operator+(const Chi_String& b) const
/* PURPOSE:    concatenate two strings
   RECEIVES:   b - the string to concatenate with self
   RETURNS:    the concatenated string
   REMARKS:    self is not affected
*/
{  int len = length();
   Chi_String r(len + b.length(), _str, len);
   if (b._str) strcpy(r._str + len, b._str);
   return r;
}

/*...........................................................................*/

Chi_String Chi_String::operator+(const char b[]) const
/* PURPOSE:    concatenate two strings
   RECEIVES:   b - the string to concatenate with self
   RETURNS:    the concatenated string
   REMARKS:    self is not affected
*/
{  int len = length();
   int blen = b ? strlen(b) : 0;
   Chi_String r(len + blen, _str, len);
   if (blen) strcpy(r._str + len, b);
   return r;
}

/*...........................................................................*/

Chi_String Chi_String::operator+(char ch) const
/* PURPOSE:    concatenate a string and a character
   RECEIVES:   ch - the character to concatenate with self
   RETURNS:    the concatenated string
   REMARKS:    self is not affected
*/
{  int len = length();
   Chi_String r(len + 1, _str, len);
   r._str[ len ] = ch;
   return r;
}

/*...........................................................................*/

int Chi_String::read_line(istream& is)
/* PURPOSE:    read a line until newline or end of file
   RECEIVES:   is - the input stream
   RETURNS:    the number of characters read
   REMARKS:    the previous contents of self is destroyed
               the terminating newline is not part of the string
*/
{  const int BUFLEN = 80;
   int len = 0;
   char buffer[BUFLEN];
   Bool more = TRUE;
   empty();
   while (more)
   {  is.get(buffer, BUFLEN - 1);
      if (is.fail()) more = FALSE;
      else
      {  int blen = strlen(buffer);
         char ch = is.get();
         if (!is.fail())
         {  if (ch == '\n') more = FALSE;
            else
            {  buffer[blen] = ch;
               buffer[blen + 1] = 0;
               blen++;
            }
         }
         len += blen;
         append(buffer);
      }
   }
   return len;
}

/*...........................................................................*/

void Chi_String::read_delimited(istream& is)
/* PURPOSE:    Read a "" delimited string (including spaces) from a stream
   RECEIVES:   is - the input stream
   REMARKS:    embedded " and \ are quoted by \
*/
{  Chi_String s;
   char ch;
   is >> ch;
   if (ch != '"') { is.clear(ios::failbit); return; }
   const int BUFSIZE = 80;
   char buffer[BUFSIZE];
   int i = 0;
   for(;;)
   {  ch = is.get();
      if (is.fail()) return; // expect closing "
      if (ch == '"') { buffer[i] = 0; s.append(buffer); self = s; return; }
      if (ch == '\\')
      {  ch = is.get(); // \ is escape for embedded "
         if (is.fail()) return; // expect escapee
      }
      buffer[i] = ch;
      i++;
      if (i >= BUFSIZE - 1) { buffer[i] = 0; s.append(buffer); i = 0; }
   }
}

//............................................................................

void Chi_String::write_delimited(ostream& os) const
/* PURPOSE:    write a "" delimited string (including spaces) to a stream
   RECEIVES:   os - the output stream
   REMARKS:    embedded " and \ are quoted by \
*/
{  os.put('"');
   char ch;
   if (_str)
      for (int i = 0; (ch = _str[i]) != 0; i++)
      {  if (ch == '"' || ch == '\\') os.put('\\');
         os.put(ch);
      }
   os.put('"');
}

//............................................................................

Chi_String Chi_String::substr(int from, int n) const
/* PURPOSE:       Get a substring from a string
   RECEIVES:      from - the starting index (0 = beginning of the string)
                  n - the number of characters to copy into substring
   RETURNS:       the substring
   REMARKS:       if from + n >= string length, takes characters until end of string
*/
{  int len = length();
   if (from < 0) from = 0;
   if (n == NPOS || from + n >= len) n = len - from;
   return Chi_String(n, _str + from, n);
}
/*...........................................................................*/

Chi_String Chi_String::field(int n) const
/* PURPOSE:    considers self as a sequence of white space terminated fields and
               returns a copy of a field
   RECEIVES:   n - the field to get (0 = first field)
   RETURNS:    the nth field
*/
{  int from, len;
   field_range(n, from, len);
   return substr(from, len);
}

/*...........................................................................*/

Chi_String Chi_String::remove_field(int n)
/* PURPOSE:    considers self as a sequence of white space terminated fields and
               removes a copy of a field
   RECEIVES:   n - the field to remove (0 = first field)
   RETURNS:    the nth field
*/
{  int from, len;
   field_range(n, from, len);
   Chi_String r = substr(from, len); // CSH 9-22-93 return without trailing spaces
   if (_str && len > 0)
   {  while (_str[from + len] && isspace(_str[from + len])) len++;
      remove(from, len);
   }
   return r;
}

/*...........................................................................*/

void Chi_String::insert_field(int n, const Chi_String& s)
/* PURPOSE:    considers self as a sequence of white space terminated fields and
               insert a new a field
   RECEIVES:   n - the field before which to insert (0 = first field)
               s - the material to insert
*/
{  if (n < 0) return;
   Chi_String f = s + " ";
   int from, len;
   field_range(n, from, len);
   insert(from, f);
}

/*...........................................................................*/

Chi_String Chi_String::replace_field(int n, const Chi_String& s)
/* PURPOSE:    considers self as a sequence of white space terminated fields and
               replaces a field with another
   RECEIVES:   n - the field to replace (0 = first field)
               s - the material to insert
*/
{  if (n < 0) return "";
   Chi_String r = remove_field(n);
   insert_field(n, s);
   return r;
}

/*.........................................................................*/

void Chi_String::detach()
/* PURPOSE:    decrement reference count and free associated object if
               necessary
*/
{  if (_str && Chi_StringHeader::header(_str)->decref() == 0)
      Chi_StringHeader::free(_str);
}

/*.........................................................................*/

void Chi_String::unique()
/* PURPOSE:    makes copy of representation if reference count > 1
*/
{  if (_str && !Chi_StringHeader::header(_str)->unique())
   {  Chi_StringHeader::header(_str)->decref();
      int len = length();
      _str = Chi_StringHeader::alloc(len, _str, len);
   }
}

/*...........................................................................*/

void Chi_String::field_range(int n, int& from, int& len) const
/* PURPOSE:    Compute the range of a field (delimited by white space)
   RECEIVES:   n - the field # (0 = first field; may be preceded by spaces)
               from (OUT) - the start of the field
               len (OUT) - the length of the field
   REMARKS:    nonexistent fields yield an empty range (from > to)
*/
{  if (!_str) { from = 0; len = NPOS; return; }
   int i = 0; // char index
   int f = 0; // field index
   while (_str[i])
   {  while (isspace(_str[i])) i++;
      if (f == n) from = i;
      while (_str[i] && !isspace(_str[i])) i++;
      if (f == n) { len = i - from; return; }
      f++;
   } 
   from = i; len = 0;   
}

/*...........................................................................*/

long Chi_String::to_long() const
/* PURPOSE:    Converts a string to a long
   RETURNS:    The equivalent long (0 if none found).
*/
{  return _str ? atol(_str) : 0;
}

/*...........................................................................*/

double Chi_String::to_double() const
/* PURPOSE:    Converts a string to double.
   RETURNS:    The equivalent double
*/
{  return _str ? atof(_str) : 0;
}

/*...........................................................................*/

void Chi_String::from_long(long n)
/* PURPOSE:    Converts a long to a string.
   RECEIVES:   The long to convert.
   REMARKS:    No domain checking of the long is performed.
*/
{  char buf[3*sizeof(long)];
   // ltoa(n, buf, 10);
   sprintf(buf, "%ld", n);
#if defined(CHI_MC21) || defined(CHI_MC40)
   self = buf;
#else
   self = Chi_String(buf);
#endif
}

/*...........................................................................*/

void Chi_String::from_double(double d)
/* PURPOSE:    Converts a double to a string (in general format, with 6 digits)
   RECEIVES:   The double to convert.
   REMARKS:    No domain checking of the double is performed.
*/
{  char buf[2 * DBL_DIG];
   // gcvt(d, 6, buf);
   sprintf(buf, "%.6g", d);
#if defined(CHI_MC21) || defined(CHI_MC40)
   self = buf;
#else
   self = Chi_String(buf);
#endif
}

/*...........................................................................*/

int Chi_String::hash(const Chi_String& s)
/* PURPOSE:    compute a hash value
   RECEIVES:   s- a string
   RETURNS:    a hash value for s
*/
{  int h = 0;
   for (int i = 0; i < s.length(); i++)
      h = 13 * h + s[i];
   return h;
}

/*...........................................................................*/

int Chi_String::compare(const Chi_String& a, const Chi_String& b) 
/* PURPOSE:    lexicographically compares two strings
   RECEIVES:   a, b - strings
   RETURNS:    < 0 if a before b
               0 if a equals b
               > 0 if a after b
*/  
{  return strcmp(a.c_str(), b.c_str());
}

/*...........................................................................*/

int Chi_String::compare_insens(const Chi_String& a, const Chi_String& b)
/* PURPOSE:    lexicographically compares two strings without regard for case
   RECEIVES:   a, b - strings
   RETURNS:    < 0 if a before b
               0 if a equals b
               > 0 if a after b
*/  
{  return chi_stricmp(a.c_str(), b.c_str());
}

/*---------------------------------------------------------------------------*/

char* Chi_StringHeader::alloc(const char s[])
/* PURPOSE:    allocate space for a header and a string
   RECEIVES:   s - the string to copy above the header
   RETURNS:    the start of the string
*/
{  if (s == 0) return 0;
   int len = strlen(s);
   if (len == 0) return 0;
   Chi_StringHeader* ph = (Chi_StringHeader*)::malloc(len + 1 + sizeof(Chi_StringHeader));
   CHI_ASSERT(ph != 0);
   ph->_length = len;
   ph->_rc = 1;
   char* t = (char*)(ph+1);
   strcpy(t, s);
   return t;
}

/*...........................................................................*/

char* Chi_StringHeader::alloc(int len, const char s[], int ncopy)
/* PURPOSE:    allocate space for a header and a string
   RECEIVES:   len - the total number of characters to allocate
               s - the string to copy above the header
               ncopy - the number of characters in s to copy
   RETURNS:    the start of the string
*/
{  if (len == 0) return 0;
   Chi_StringHeader* ph = (Chi_StringHeader*)::malloc(len + 1 + sizeof(Chi_StringHeader));
   CHI_ASSERT(ph != 0);
   ph->_length = len;
   ph->_rc = 1;
   char* t = (char*)(ph+1);
   CHI_ASSERT(len >= ncopy);
   if (s && ncopy) strncpy(t, s, ncopy);
   t[ len ] = 0;
   return t;
}

/*...........................................................................*/

void Chi_StringHeader::free(char* s)
/* PURPOSE:    free a block containing a header and string
   RECEIVES:   s - the start of the string
*/
{  Chi_StringHeader* ph = header(s);
   CHI_ASSERT(ph->_rc == 0);
   CHI_ASSERT(ph->_length > 0);
   ::free(ph);
}

/* TEST ********************************************************************/

#ifdef CHISTRNG_TEST

int main()
{  Chi_String s("Hello, World");
   Chi_String t;
   s = s + t;
   Chi_String u = s;
   Chi_String a = u.remove(0, 4);
   Chi_String b = a;
   b.to_upper();
   Chi_String c = b;

   cout << s.substr(0, 5) << s[5] << s.substr(6) << endl;
   cout << a << u << endl;
   if (s.substr(0, 5) < s.substr(7))
      cout << c.to_lower() << endl << c << endl;

   s = "c:/borlandc/include/iostream.h";

   cout << s.substr(0, s.find_first_of("abc:/", 10) + 1) << endl;
   cout << s.substr(s.find_first_not_of("abc:/") - 1) << endl;

   cout << "Here we test the bounds error checking." << endl;
   char ch = s[100];
   s[100] = ch;
   return 0;
}

#endif

/***************************************************************************/
