/****************************************************************************
** COPYRIGHT (C):    1994 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          Mastering Object-Oriented Design
** FILE:             employee.cpp
** PURPOSE:          Employee and manager class
** VERSION           1.0
** LANGUAGE:         Borland C++ 3.0
** TARGET:           IBM PC
** PROGRAMMER:       Cay Horstmann (CSH)
** START DATE:       4-30-92 (CSH)
** UPDATE HISTORY:
****************************************************************************/

/* IMPORT ******************************************************************/

#ifndef EXPORT
#include "chisetup.h"
#endif

EXPORT
#include <iostream.h>

EXPORT
#include "chistrng.h"
EXPORT
#include "chidate.h"
EXPORT
#include "chiarray.h"

#include "employee.h"

/* TYPES *******************************************************************/

#ifndef EXPORT

class Employee
/* An employee in a company--used in many examples
*/
{
public:
   Employee();
   Employee(Chi_String);
   Employee(Chi_String, int id);
   Employee(Chi_String n, int hd, int hm, int hy);

   void set_id(int id);
   void set_salary(double s);
   void raise_salary(double p);
   int compare(Employee) const;
   virtual void print() const;
   virtual void print(ostream& os) const;
   int id() const;
   double hourly_wage() const;
   virtual double weekly_pay(double hours) const;

   virtual ~Employee() {}

private:
   Chi_String _name;
   double _wage;
   int _id;
   Chi_Date _hiredate;
};

/*-------------------------------------------------------------------------*/

class Manager : public Employee
/* PURPOSE:    A manager supervising employees
*/
{
public:
   Manager();
   Manager(Chi_String s);

   void add(Employee* e);
   virtual void print(ostream&) const;
   virtual double weekly_pay(double hours) const;
private:
   Chi_PtrArray<Employee> _superv; // ID numbers of supervised employees
};

#endif

/* FUNCTIONS ***************************************************************/

Employee::Employee()
: _wage(0),
  _id(0)
{}

/*.........................................................................*/

Employee::Employee(Chi_String n)
/* RECEIVES:   n - name
*/
:  _name(n),
   _wage(0),
   _id(0)
{}

/*.........................................................................*/

Employee::Employee(Chi_String n, int hd, int hm, int hy)
/* RECEIVES:   n - name
               hd, hm, hy - hire day, month, year
*/
:  _name(n),
   _wage(0),
   _id(0),
   _hiredate(hd, hm, hy)
{}

/*.........................................................................*/

Employee::Employee(Chi_String n, int id)
/* RECEIVES:   n - name
               id - ID number
*/
:  _name(n),
   _wage(0),
   _id(id)
{}

/*.........................................................................*/

void Employee::set_salary(double s)
/* PURPOSE:    set employee salary
   RECEIVES:   s - new salary
*/
{  _wage = s;
}

/*.........................................................................*/

void Employee::raise_salary(double p)
/* PURPOSE:    raise employee salary
   RECEIVES:   p - the percentage to raise by
*/
{  _wage *= 1 + p;
}

/*.........................................................................*/

double Employee::weekly_pay(double h) const
/* PURPOSE:    computes pay for a week
   RECEIVES:   h - number of hours worked in that week
   RETURNS:    employee pay
   REMARKS:    overtime is paid at 1.5 * the normal rate
*/
{  return (h + chi_max(h - 40, 0.0) * 0.5) * hourly_wage();
}

/*.........................................................................*/

void Employee::set_id(int i)
/* PURPOSE:    set employee id
   RECEIVES:   i - new ID number
*/
{  _id = i;
}

/*.........................................................................*/

int Employee::id() const
/* RETURNS:    employee ID number
*/
{  return _id;
}

/*.........................................................................*/

double Employee::hourly_wage() const
/* RETURNS:    pay for one hour
*/
{  const int WORKING_HRS_PER_YEAR = 2000;
      // this is a commonly used figure -- 50 weeks at 40 hours/week
   return _wage / WORKING_HRS_PER_YEAR;
}

/*.........................................................................*/

void Employee::print() const
/* PURPOSE:    print employee data to standard output
   REMARKS:    the call print(cout) is a virtual call
*/
{  print(cout);
}

/*.........................................................................*/

void Employee::print(ostream& os) const
/* PURPOSE:    print employee data to a stream
   RECEIVES:   os - the stream
*/
{  os << _name << " " << _id << " " << _wage << " " << _hiredate << endl;
}

/*.........................................................................*/

int Employee::compare(Employee b) const
/* PURPOSE:    compare two employee records for sorting
   RECEIVES:   b - the employee to compare with self
   RETURNS:    < 0 if self comes before b, 0 if they have the same id, >0
               otherwise
*/
{  return _id - b._id;
}

/*-------------------------------------------------------------------------*/

Manager::Manager(Chi_String name)
/* RECEIVES:   name - manager name
*/
: Employee(name)
{}


/*.........................................................................*/

void Manager::add(Employee* e)
/* PURPOSE:    add an employee to be supervised by this manager
   RECEIVES:   e - the employee to be supervised
*/
{  _superv.append(e);
}

/*.........................................................................*/

void Manager::print(ostream& os) const
/* PURPOSE:    print employee data to a stream
   RECEIVES:   os - the stream
*/
{  Employee::print(os);

   for (int i = _superv.low(); i <= _superv.high(); i++)
   {  os << "  ";
      _superv[i]->print(os);
   }
}

/*.........................................................................*/

double Manager::weekly_pay(double h) const
/* PURPOSE:    computes pay for a week
   RECEIVES:   h - number of hours worked in that week
   RETURNS:    employee pay
   REMARKS:    no overtime for managers
*/
{  double w = hourly_wage();
   if (h <= 30) return h * w;
   else return 40 * w;
}

/* TEST ********************************************************************/

#if defined(EMPLOYEE_TEST) || defined(employee_TEST)

int employee_comp(const Employee* e, const Employee* f)
/* PURPOSE:    compare two employee records for sorting
   RECEIVES:   e, f - pointers to employees
   RETURNS:    < 0 if e comes before f, 0 if they have the same id, >0
               otherwise
*/
{  return (*e).compare(*f);
}

/*.........................................................................*/

double salary(const Employee& e, Chi_Date from, Chi_Date to)
/* PURPOSE:    Compute employee salary in a date range
   RECEIVES:   e - the employee whose salary to compute
               from, to - the date range
   RETURNS:    the total salary earned
   REMARKS:    We assume 8 hours of work for every weekday in the pay period
   EXERCISE:   This computation is very inefficient. Speed it up by doing
               a week at a time, taking care of the partial weeks at the
               beginning and the end.
*/
{  double s = 0;
   double h = 0;
   for (Chi_Date d = from; d <= to; d.advance(1))
   {  Chi_Date::Weekday w = d.weekday();
      if (w != Chi_Date::SAT && w != Chi_Date::SUN)
      {  h += 8;
         if (w == Chi_Date::FRI)
         {  s += e.weekly_pay(h);
            h = 0;
         }
      }
   }
   if (h > 0) s += e.weekly_pay(h);
   return s;
}

/*.........................................................................*/

int main()
{  Chi_PtrArray<Employee> staff(1, 5);

   Employee* harry = new Employee("Harry Hacker", 149);
   Manager* carol = new Manager("Carol Smith");

   harry->set_salary(40000);
   harry->raise_salary(0.08);
   carol->set_salary(60000);
   carol->raise_salary(0.03);

   staff[1] = new Employee("Joe User", 16);
   staff[2] = new Employee("Karl Schiller", 2536);
   staff[3] = new Employee("Jessica Chang", 31, 3, 1961);

   for (int i = 1; i <= 3; i++) staff[i]->set_salary(35000);
   staff[3]->set_id(3);

   staff[4] = harry;
   staff[5] = carol;

   carol->add(harry);
   carol->add(staff[3]);

   staff.qsort(employee_comp);

   cout.precision(2);
   cout.setf(ios::fixed | ios::showpoint);

   for (i = staff.low(); i <= staff.high(); i++)
      staff[i]->print();
   for (i = staff.low(); i <= staff.high(); i++)
      cout << staff[i]->id() << " : " << staff[i]->weekly_pay(48) << endl;

   Chi_Date from(1, 1, 1994);
   Chi_Date to(31, 3, 1994);

   for (i = staff.low(); i <= staff.high(); i++)
      cout << staff[i]->id() << " : " << salary(*staff[i], from, to) << endl;
   for (i = staff.low(); i <= staff.high(); i++)
      delete staff[i];

   return 0;
}

#endif
