/****************************************************************************
** COPYRIGHT (C):    1994 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          Mastering Object-Oriented Design
** FILE:             gedfrmwk.cpp
** PURPOSE:          Graphics editor framework
** VERSION           1.0
** LANGUAGE:         Borland C++ 3.1
** TARGET:           IBM PC
** PROGRAMMERS:      Cay S. Horstmann (CSH)
** START DATE:       3/21/94
** UPDATE HISTORY:
****************************************************************************/

/* IMPORT ******************************************************************/

#ifndef EXPORT
#include "chisetup.h"
#endif

#include <fstream.h>

EXPORT
#include "chiarray.h"
EXPORT
#include "chistrng.h"
EXPORT
#include "chishape.h"
#include "chierror.h"
#include "gedfrmwk.h"

/* TYPES *******************************************************************/

#ifndef EXPORT

class Node
/* PURPOSE:    An abstract class for a node in a graph
*/
{
public:
   virtual Chi_String name() const = 0;
   virtual void plot(Chi_GraphicsContext& gc) const = 0;
   virtual void plot_icon(Chi_Rectangle r, Chi_GraphicsContext& gc) const = 0;
   virtual Node* clone() const = 0;
   virtual Chi_Rectangle enclosing_rect() const = 0;
   virtual Bool is_inside(Chi_Point p) const = 0;
   virtual Chi_Point boundary_point(Chi_Point exterior) const = 0;
   virtual void read(istream& is) = 0;
   virtual void print(ostream& os) const  = 0;

   void move(Chi_Point p);
   Chi_Point center() const;

private:
   Chi_Point _center;
};

/*-------------------------------------------------------------------------*/

class Edge
/* PURPOSE:    An abstract class for an edge joining two nodes
*/
{
public:
   void connect(Node* f, Node* t);

   virtual Chi_String name() const = 0;
   virtual void plot(Chi_GraphicsContext& gc) const = 0;
   virtual void plot_icon(Chi_Rectangle r, Chi_GraphicsContext& gc) const = 0;
   virtual Edge* clone() const = 0;
   virtual Bool is_on(Chi_Point p) const = 0;
   virtual void read(istream& is) = 0;
   virtual void print(ostream& os) const = 0;

   const Node* from() const;
   const Node* to() const;

private:
   Node* _from;
   Node* _to;
};

/*-------------------------------------------------------------------------*/

class Diagram
/* PURPOSE:    A collection of nodes and edges
*/
{
public:
   void add_node(Node*);
   void add_edge(Edge*);
   void remove(Chi_Point);
   Node* find_node(Chi_Point);
   int find_node(const Node*) const;
   Edge* find_edge(Chi_Point) const;
   int search_node(Chi_Point p) const;
   void plot(Chi_GraphicsContext&) const;
   void print(ostream& os) const;
private:
   Chi_PtrArray<Node> _nodes;
   Chi_PtrArray<Edge> _edges;
};

/*-------------------------------------------------------------------------*/

class DiagramEditor
/* PURPOSE:    An editor for diagrams made from nodes and edges
*/
{
public:
   DiagramEditor();
   enum Mode { EXIT, MOVE, ERASE, NODE, EDGE };

   virtual void register_types() = 0;
   void run(Chi_String infile);

   void register_node(Node*);
   void register_edge(Edge*);

   void read(istream& is);
   Node* make_node(Chi_String) const;
   Edge* make_edge(Chi_String) const;

   void event_loop();
   void rubber_band();
   void display();

   void plot_buttons();
   void plot_exit_button(Chi_Rectangle);
   void plot_move_button(Chi_Rectangle);
   void plot_erase_button(Chi_Rectangle);
   void mark_current_button();
   Node* clone_current_node();
   Edge* clone_current_edge();
   Mode click_button(Chi_Point p);
   Mode mode() const;

   int mouse_handler(const Chi_MouseEvent& e);
   static int mouse_handler(const Chi_MouseEvent& e, void* p);

private:

   Chi_PtrArray<Node> _node_types;
   Chi_PtrArray<Edge> _edge_types;

   Diagram _diagram;

   Chi_Rectangle _button_rect;
   int _current_button;

   Node* _start_node;
   Chi_Point _start_point;
   Chi_Point _current_point;

   Chi_GraphicsContext _gc;
   Chi_String _filename;
};

#endif

/* FUNCTIONS ***************************************************************/

DiagramEditor::DiagramEditor()
:  _start_node(0)
{
}

/*.........................................................................*/

void DiagramEditor::read(istream& is)
/* PURPOSE:    read a diagram
   RECEIVES:   is - input stream
   REMARKS:    Each entry has the format
                  nodetag number nodeprops
               or
                  edgetag from to edgeprops
*/
{  Chi_PtrArray<Node> nodes;
   while (!is.fail())
   {  Chi_String tag;
      is >> tag;
      if (is.fail()) return;
      Node* n = make_node(tag);
      if (n)
      {  int nn;
         is >> nn;
         n->read(is);
         if (!is.fail())
         {  _diagram.add_node(n);
            nodes.set(nn, n); // remember node number for edges
         }
      }
      else
      {  Edge* e = make_edge(tag);
         if (e)
         {  int f, t;
            is >> f >> t;
            e->read(is);
            if (!is.fail())
            {  e->connect(nodes.get(f), nodes.get(t));
               _diagram.add_edge(e);
            }
         }
         else
         {  is.clear(ios::failbit);
            return;
         }
      }
   }
}

/*.........................................................................*/

void DiagramEditor::register_node(Node* n)
/* PURPOSE:    register a node type with the editor
   RECEIVES:   n - a sample instance of the derived type
*/
{  _node_types.append(n);
}

/*.........................................................................*/

void DiagramEditor::register_edge(Edge* e)
/* PURPOSE:    register an edge type with the editor
   RECEIVES:   e - a sample instance of the derived type
*/
{  _edge_types.append(e);
}

/*.........................................................................*/

Node* DiagramEditor::make_node(Chi_String s) const
/* PURPOSE:    make a new node
   RECEIVES:   s - the name of the node type
   RETURNS:    a pointer to the new node allocated on the heap
*/
{  for (int i = _node_types.low(); i <= _node_types.high(); i++)
      if (_node_types.get(i)->name() == s)
         return _node_types.get(i)->clone();
   return 0;
}

/*.........................................................................*/

Edge* DiagramEditor::make_edge(Chi_String s) const
/* PURPOSE:    make a new edge
   RECEIVES:   s - the name of the edge type
   RETURNS:    a pointer to the new edge allocated on the heap
*/
{  for (int i = _edge_types.low(); i <= _edge_types.high(); i++)
      if (_edge_types.get(i)->name() == s)
         return _edge_types.get(i)->clone();
   return 0;
}

/*.........................................................................*/

void DiagramEditor::rubber_band()
/* PURPOSE:    draw a rubber band rectange or line from the start position
               of the mouse move to the current position
*/
{  if (_start_node)
   {  if (mode() == MOVE)
      {  Chi_Rectangle r = _start_node->enclosing_rect();
         r.move(_current_point.x() - _start_point.x(),
            _current_point.y() - _start_point.y());
         r.plot(_gc);
      }
      else if (mode() == EDGE)
      {  Chi_Segment s(_start_point, _current_point);
         s.plot(_gc);
      }
   }
}

/*.........................................................................*/

void DiagramEditor::display()
/* PURPOSE:    display the current diagram
*/
{  // erase screen
   _gc.clear();
   _diagram.plot(_gc);
}

/*.........................................................................*/

void DiagramEditor::run(Chi_String infile)
/* PURPOSE:    run the editor
   RECEIVES:   infile - the name of the file from which the diagram is read,
                  and to which it is written when the editor quits
*/
{  register_types(); // supplied by derived class

   plot_buttons();
   _current_button = MOVE;
   mark_current_button();
   _filename = infile;
   ifstream is(infile.c_str());
   read(is);
   is.close();
   display();
   _gc.mouse_events(mouse_handler, this);
}

/*.........................................................................*/

void DiagramEditor::plot_exit_button(Chi_Rectangle r)
/* PURPOSE:    plot the exit button
   RECEIVES:   r - enclosing rectangle
*/
{  _gc.set_font_style(Chi_GraphicsContext::SANS_FONT);
   Chi_ScalableText(r.left_top(), "Exit", 0.35).plot(_gc);
}

/*.........................................................................*/

void DiagramEditor::plot_move_button(Chi_Rectangle r)
/* PURPOSE:    plot the move button
   RECEIVES:   r - enclosing rectangle
*/
{  _gc.set_font_style(Chi_GraphicsContext::SANS_FONT);
   Chi_ScalableText(r.left_top(), "Move", 0.35).plot(_gc);
}

/*.........................................................................*/

void DiagramEditor::plot_erase_button(Chi_Rectangle r)
/* PURPOSE:    plot the erase button
   RECEIVES:   r - enclosing rectangle
*/
{  _gc.set_font_style(Chi_GraphicsContext::SANS_FONT);
   Chi_ScalableText(r.left_top(), "Erase", 0.35).plot(_gc);
}

/*.........................................................................*/

void DiagramEditor::plot_buttons()
/* PURPOSE:    plot all buttons in the left hand side of the display
   REMARKS:    sets _button_rect
*/
{  Chi_Rectangle r(Chi_Point(0.08, 0.08), Chi_Point(0.92, 0.92));
   r.plot(_gc);
   plot_exit_button(r);
   r.move(0, 1);
   r.plot(_gc);
   plot_move_button(r);
   r.move(0, 1);
   r.plot(_gc);
   plot_erase_button(r);
   for (int i = _node_types.low(); i <= _node_types.high(); i++)
   {  r.move(0, 1);
      r.plot(_gc);
      _node_types.get(i)->plot_icon(r, _gc);
   }
   for (i = _edge_types.low(); i <= _edge_types.high(); i++)
   {  r.move(0, 1);
      r.plot(_gc);
      _edge_types.get(i)->plot_icon(r, _gc);
   }
#ifndef CHI_BC45
   _button_rect = Chi_Rectangle(Chi_Point(0, 0), Chi_Point(1, r.ybottom()));
#else
   // code above causes internal error in BC45
   Chi_Point p2 = Chi_Point(1, r.ybottom());
   _button_rect = Chi_Rectangle(Chi_Point(0, 0), p2);
#endif
}

/*.........................................................................*/

void DiagramEditor::mark_current_button()
/* PURPOSE:    mark the currently selected button
*/
{  // set the clip rectangle to the entire screen
   const int XRIGHT = 20;
   const int YBOTTOM = 15;
   _gc.set_clip_rectangle(0, 0, XRIGHT, YBOTTOM);
   _gc.set_rop_mode(Chi_GraphicsContext::XOR_ROP);
   // draw outline
   Chi_Rectangle r(Chi_Point(0.04, _current_button+0.04),
      Chi_Point(0.97, _current_button + 0.97));
   r.plot(_gc);
   _gc.set_rop_mode(Chi_GraphicsContext::COPY_ROP);
   // set the clip rectangle to protect the button column
   _gc.set_clip_rectangle(1, 0, XRIGHT, YBOTTOM);
}

/*.........................................................................*/

DiagramEditor::Mode DiagramEditor::click_button(Chi_Point p)
/* PURPOSE:    process a mouse click in the button area
   RECEIVES:   p - the location of the mouse click
   RETURNS:    the new draw mode
*/
{  mark_current_button();
   _current_button = (int)p.y();
   mark_current_button();
   return mode();
}

/*.........................................................................*/

DiagramEditor::Mode DiagramEditor::mode() const
/* RETURNS:    the current draw mode
*/
{  if (_current_button < NODE) return (Mode)_current_button;
   int i = _current_button - NODE + _node_types.low();
   if (_node_types.low() <= i && i <= _node_types.high()) return NODE;
   i = _current_button - NODE - _node_types.high() - 1 + _edge_types.low();
   if (_edge_types.low() <= i && i <= _edge_types.high()) return EDGE;
   return EXIT; // shouldn't happen
}

/*.........................................................................*/

Node* DiagramEditor::clone_current_node()
/* PURPOSE:    make a clone of the currently active node
   RETURNS:    a pointer to a new node allocated on the heap
*/
{  int i = _current_button - NODE + _node_types.low();
   Node* n = _node_types.get(i);
   return n->clone();
}

/*.........................................................................*/

Edge* DiagramEditor::clone_current_edge()
/* PURPOSE:    make a clone of the currently active edge
   RETURNS:    a pointer to a new edge allocated on the heap
*/
{  int i = _current_button - NODE - 1 - _node_types.high() + _edge_types.low();
   Edge* e = _edge_types.get(i);
   return e->clone();
}

/*.........................................................................*/

int DiagramEditor::mouse_handler(const Chi_MouseEvent& e, void* p)
/* PURPOSE:    handle mouse events
   RECEIVES:   e - the mouse event
               p - a pointer to the diagram editor
   RETURNS:    0 to exit, any other value to contine
*/
{  return ((DiagramEditor*)p)->mouse_handler(e);
}

/*.........................................................................*/

int DiagramEditor::mouse_handler(const Chi_MouseEvent& e)
/* PURPOSE:    handle mouse events
   RECEIVES:   e - the mouse event
   RETURNS:    0 to exit, any other value to contine
*/
{  Mode m = mode();
   switch(e.type())
   {  case Chi_MouseEvent::LBUTTONDOWN:
         _gc.set_rop_mode(Chi_GraphicsContext::XOR_ROP);
         _current_point = Chi_Point(e.x(), e.y());
         _start_point = _current_point;
         if (m == MOVE || m == EDGE)
         {  _start_node = _diagram.find_node(_start_point);
            if (_start_node)
               _start_point = _start_node->center();
         }
         rubber_band();
         break;

      case Chi_MouseEvent::LBUTTONUP:
         rubber_band();
         _gc.set_rop_mode(Chi_GraphicsContext::COPY_ROP);
         _current_point = Chi_Point(e.x(), e.y());
#ifndef CHI_WC10
         if (_start_point.distance(_current_point) < 0.05)
#else
         // normal code causes internal error in Watcom 10.0
         double dist;
         dist = _start_point.distance(_current_point);
         if (dist < 0.05)
#endif
         {  if (_button_rect.is_inside(_current_point))
                     {  if (click_button(_current_point) == EXIT)
               {  _diagram.print(ofstream(_filename.c_str()));
                  _gc.close(/* wait = */ FALSE);
                  return 0;
               }
            }
            else if (m == NODE)
               // add new node of current type
            {  Node* n = clone_current_node();
               n->move(_current_point);
               _diagram.add_node(n);
               display();
            }
            else if (m == ERASE)
               // erase node under the cursor
            {  _diagram.remove(_current_point);
               display();
            }
         }
         else if (_start_node)
         {  if (m == EDGE)
               // connect startnode, endnode
            {  Node* n = _diagram.find_node(_current_point);
               if (n && n != _start_node)
               {  Edge* e = clone_current_edge();
                  e->connect(_start_node, n);
                  _diagram.add_edge(e);
                  display();
               }
            }
            else if (m == MOVE)
            {  _start_node->move(_current_point);
               display();
            }
         }
         _start_node = 0;
         break;

      case Chi_MouseEvent::MOUSEMOVE:
         if (_start_node)
         {  rubber_band();
            _current_point = Chi_Point(e.x(), e.y());
            rubber_band();
         }
         break;
   }
   return 1;
}

/*-------------------------------------------------------------------------*/

void Diagram::add_node(Node* n)
/* PURPOSE:    add a node to the diagram
   RECEIVES:   n - the node to add
*/
{  _nodes.append(n);
}

/*.........................................................................*/

void Diagram::add_edge(Edge* n)
/* PURPOSE:    add an edge to the diagram
   RECEIVES:   n - the edge to add
*/
{  _edges.append(n);
}

/*.........................................................................*/

void Diagram::print(ostream& os) const
/* PURPOSE:    write the diagram to a stream
   RECEIVES:   os - the output stream
*/
{  for (int i = _nodes.low(); i <= _nodes.high(); i++)
   {  Node* n = _nodes.get(i);
      os << n->name() << " " << i << " ";
      n->print(os);
      os << endl;
   }
   for (i = _edges.low(); i <= _edges.high(); i++)
   {  Edge* e = _edges.get(i);
      int f = find_node(e->from());
      int t = find_node(e->to());
      os << e->name() << " " << f << " " << t << " ";
      e->print(os);
      os << endl;
   }
}

/*.........................................................................*/

void Diagram::plot(Chi_GraphicsContext& gc) const
/* PURPOSE:    plot the diagram
   RECEIVES:   gc - the graphics context
*/
{  for (int i = _edges.low(); i <= _edges.high(); i++)
   {  Edge* e = _edges.get(i);
      e->plot(gc);
   }
   for (i = _nodes.low(); i <= _nodes.high(); i++)
   {  Node* n = _nodes.get(i);
      n->plot(gc);
   }
}

/*.........................................................................*/

Node* Diagram::find_node(Chi_Point p)
/* PURPOSE:    find the node containing a point
   RECEIVES:   p - a point
   RETURNS:    a node in the diagram that has p in its inside
*/
{  int n = search_node(p);
   if (n >= _nodes.low()) return _nodes.get(n); else return 0;
}

/*.........................................................................*/

int Diagram::search_node(Chi_Point p) const
/* PURPOSE:    find the node containing a point
   RECEIVES:   p - a point
   RETURNS:    the offset in the _nodes array of a node that has p in its
               inside, low - 1 if none found
*/
{  for (int i = _nodes.low(); i <= _nodes.high(); i++)
      if (_nodes.get(i)->is_inside(p)) return i;
   return _nodes.low() - 1;
}

/*.........................................................................*/

void Diagram::remove(Chi_Point p)
/* PURPOSE:    Erase feature under point p
*/
{  for (int i = _edges.low(); i <= _edges.high(); i++)
   {  if (_edges.get(i)->is_on(p))
      {  delete _edges.remove(i);
         return;
      }
   }

   i = search_node(p);
   if (i >= _nodes.low()) // erase that node and all incident edges
   {  Node* n = _nodes.remove(i);
      for (int i = _edges.low(); i <= _edges.high(); i++)
      {  Edge* e = _edges.get(i);
         if (e->from() == n || e->to() == n)
         {  delete _edges.remove(i);
            i--;
         }
      }
      delete n;
   }
}

/*.........................................................................*/

int Diagram::find_node(const Node* n) const
/* PURPOSE:    find the offset in the _nodes array of a node
   RECEIVES:   n - a node
   RETURNS:    the offset in the _nodes array of n
*/
{  for (int i = _nodes.low(); i <= _nodes.high(); i++)
      if (_nodes.get(i) == n) return i;
   return _nodes.low() - 1;
}

/*-------------------------------------------------------------------------*/

Chi_Point Node::center() const
/* RETURNS:    the center of the node
*/
{  return _center;
}

void Node::move(Chi_Point p)
/* PURPOSE:    move a node
   RECEIVES:   p - new center point
*/
{  _center = p;
}

/*-------------------------------------------------------------------------*/

const Node* Edge::from() const
/* RETURNS:    the starting node of the edge
*/
{  return _from;
}

/*.........................................................................*/

const Node* Edge::to() const
/* RETURNS:    the ending node of the edge
*/
{  return _to;
}

/*.........................................................................*/

void Edge::connect(Node* f, Node* t)
/* PURPOSE:    connect this edge to nodes
   RECEIVES:   f, t - starting, ending node
*/
{  CHI_ASSERT(f != 0 && t != 0);
   _from = f;
   _to = t;
}

/***************************************************************************/
