/****************************************************************************
** COPYRIGHT (C):    1994 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          Mastering Object-Oriented Design
** FILE:             graphed1.cpp
** PURPOSE:          A simple graphics editor made from the framework
** VERSION           1.0
** LANGUAGE:         Borland C++ 3.1
** TARGET:           IBM PC
** PROGRAMMERS:      Cay S. Horstmann (CSH)
** START DATE:       3/21/94
** UPDATE HISTORY:
****************************************************************************/

/* IMPORT ******************************************************************/

#include <math.h>

#include "gedfrmwk.h"

/* TYPES *******************************************************************/

class CircleNode : public Node
/* PURPOSE:    a node shaped like a circle
*/
{
public:
   virtual Chi_String name() const;
   virtual void plot(Chi_GraphicsContext& gc) const;
   virtual void plot_icon(Chi_Rectangle r, Chi_GraphicsContext& gc) const;
   virtual Node* clone() const;
   virtual Chi_Rectangle enclosing_rect() const;
   virtual Bool is_inside(Chi_Point p) const;
   virtual Chi_Point boundary_point(Chi_Point exterior) const;
   virtual void read(istream& is);
   virtual void print(ostream& os) const;
};

/*-------------------------------------------------------------------------*/

class Connection : public Edge
/* PURPOSE:    a straight line connecting two circles
*/
{
public:
   virtual Chi_String name() const;
   virtual void plot(Chi_GraphicsContext& gc) const;
   virtual void plot_icon(Chi_Rectangle r, Chi_GraphicsContext& gc) const;
   virtual Edge* clone() const;
   virtual Bool is_on(Chi_Point p) const;
   virtual void read(istream& is);
   virtual void print(ostream& os) const;
};

/*-------------------------------------------------------------------------*/

class SimpleEditor : public DiagramEditor
/* PURPOSE:    an editor for diagrams made up of circles and connections
*/
{
public:
   void register_types();
};

/* GLOBALS *****************************************************************/

const double RADIUS = 0.5;
   // the radius of a circle
const double ICON_RAD = 0.33;
   // the radius of a circle in the node icon

/* FUNCTIONS ***************************************************************/

int chi_main(int argc, char* argv[])
{  SimpleEditor app;
   if (argc == 2)
      app.run(argv[1]);
   else
      app.run("UNTITLED.GED");
   return 0;
};

/*-------------------------------------------------------------------------*//*-------------------------------------------------------------------------*//*-------------------------------------------------------------------------*//*-------------------------------------------------------------------------*/

void SimpleEditor::register_types()
/* PURPOSE:    register the node and edge types used in the diagrams
*/
{  register_node(new CircleNode);
   register_edge(new Connection);
}

/*-------------------------------------------------------------------------*//*-------------------------------------------------------------------------*//*-------------------------------------------------------------------------*//*-------------------------------------------------------------------------*/

Chi_String CircleNode::name() const
/* RETURNS:    the type name of this node
*/
{  return "CircleNode";
}

/*.........................................................................*/

void CircleNode::plot(Chi_GraphicsContext& gc) const
/* PURPOSE:    plots this node
   RECEIVES:   gc - graphics context
*/
{  Chi_Ellipse(center(), RADIUS, RADIUS).plot(gc);
}

/*.........................................................................*/

void CircleNode::plot_icon(Chi_Rectangle r, Chi_GraphicsContext& gc) const
/* PURPOSE:    plots the icon for this node
   RECEIVES:   r - enclosing rectangle
               gc - graphics context
*/
{  Chi_Ellipse(r.center(), ICON_RAD, ICON_RAD).plot(gc);
}

/*.........................................................................*/

Node* CircleNode::clone() const
/* RETURNS:    a new node on the heap that is a copy of this node
*/
{  return new CircleNode(self);
}

/*.........................................................................*/

Chi_Rectangle CircleNode::enclosing_rect() const
/* RETURNS:    the rectangle enclosing this node
*/
{  Chi_Point c = center();
   Chi_Point d = c;
   c.move(-RADIUS, -RADIUS);
   d.move(RADIUS, RADIUS);
   return Chi_Rectangle(c, d);
}

/*.........................................................................*/

Bool CircleNode::is_inside(Chi_Point p) const
/* PURPOSE:    test whether a point is contained in this node
   RECEIVES:   p - a point
   RETURNS:    TRUE if p is in the interior of this node
*/
{  return Chi_Ellipse(center(), RADIUS, RADIUS).is_inside(p);
}

/*.........................................................................*/

Chi_Point CircleNode::boundary_point(Chi_Point exterior) const
/* RETURNS:    a point on the boundary of this node in a given direction
   RECEIVES:   exterior - a point outside this node
   RETURNS:    the point of intersection of the boundary and the line
               joining the center with the given exterior point
*/
{  Chi_Point r = center();
   double a = r.angle(exterior);
   r.move(RADIUS * cos(a), RADIUS * sin(a));
   return r;
}

/*.........................................................................*/

void CircleNode::read(istream& is)
/* PURPOSE:    read this node's attributes from a stream
   RECEIVES:   is - an input stream
*/
{  Chi_Point p; p.read(is); move(p);
}

/*.........................................................................*/

void CircleNode::print(ostream& os) const
/* PURPOSE:    write this node's attributes to a stream
   RECEIVES:   os - an output stream
*/
{  center().print(os);
}

/*-------------------------------------------------------------------------*//*-------------------------------------------------------------------------*//*-------------------------------------------------------------------------*//*-------------------------------------------------------------------------*/

Chi_String Connection::name() const
/* RETURNS:    the type name of this edge
*/
{  return "Connection";
}

/*.........................................................................*/

void Connection::plot(Chi_GraphicsContext& gc) const
/* PURPOSE:    plots this edge
   RECEIVES:   gc - graphics context
*/
{  Chi_Point a = from()->boundary_point(to()->center());
   Chi_Point b = to()->boundary_point(from()->center());
   Chi_Segment(a, b).plot(gc);
}

/*.........................................................................*/

void Connection::plot_icon(Chi_Rectangle r, Chi_GraphicsContext& gc) const
/* PURPOSE:    plots the icon for this edge
   RECEIVES:   r - enclosing rectangle
               gc - graphics context
*/
{  Chi_Point c = r.center();
   Chi_Point d = c;
   c.move(-0.33, -0.33);
   d.move(0.33, 0.33);
   Chi_Segment(c, d).plot(gc);
}

/*.........................................................................*/

Edge* Connection::clone() const
/* RETURNS:    a new edge on the heap that is a copy of this edge
*/
{  return new Connection(self);
}

/*.........................................................................*/

Bool Connection::is_on(Chi_Point p) const
/* PURPOSE:    test whether a point lies on this edge
   RECEIVES:   p - a point
   RETURNS:    TRUE if p lies on (or is very close to) this edge
*/
{  Chi_Point a = from()->boundary_point(to()->center());
   Chi_Point b = to()->boundary_point(from()->center());
   return Chi_Segment(a, b).distance(p) < 0.1;
}

/*.........................................................................*/

void Connection::read(istream&)
/* PURPOSE:    read this node's attributes from a stream
   RECEIVES:   is - an input stream
*/
{}

/*.........................................................................*/

void Connection::print(ostream& ) const
/* PURPOSE:    write this node's attributes to a stream
   RECEIVES:   os - an output stream
*/
{}

/***************************************************************************/
