#ifndef PSTREAM_H
#define PSTREAM_H

/****************************************************************************
** COPYRIGHT (C):    1994 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          Mastering Object-Oriented Design
** FILE:             pstream.h
** PURPOSE:          A simple persistent stream implementation
** VERSION           1.0
** LANGUAGE:         Borland C++ 3.1
** TARGET:           IBM PC
** PROGRAMMERS:      Cay S. Horstmann (CSH)
** START DATE:       September 14, 1994
** UPDATE HISTORY:
****************************************************************************/

#line 16 "PSTREAM.cpp"

#include "chisetup.h"

#line 20 "PSTREAM.cpp"

#include <fstream.h>

#line 22 "PSTREAM.cpp"

#ifndef CHISTRNG_H
#include "chistrng.h"
#endif

#line 24 "PSTREAM.cpp"

#ifndef CHIARRAY_H
#include "chiarray.h"
#endif

#line 26 "PSTREAM.cpp"

#ifndef CHIMAP_H
#include "chimap.h"
#endif

#line 33 "PSTREAM.cpp"


class Persistent;

class PersistentStream : public fstream
/* PURPOSE:    a stream for saving and loading persistent objects
*/
{
public:
   PersistentStream();
   void load(Chi_String);
   void save(Chi_String);
   void close();
   Bool is_saving() const;

   void archive(Persistent*& x);
   Persistent* find(int) const;
   void add(int, Persistent*);
   int find(const Persistent* p) const;
   int add(const Persistent* p);

private:
   enum Mode { CLOSED, LOAD, SAVE };
   Chi_PtrArray<Persistent> _toptr;
   Chi_Map<const Persistent*, int> _fromptr;
   int _maxnum;
   Mode _mode;
};

/*-------------------------------------------------------------------------*/

class Persistent
/* PURPOSE:    the base class for persistent objects
*/
{
public:
   virtual void archive(PersistentStream&) = 0;
   virtual Chi_String name() const = 0;
   virtual Persistent* clone() const = 0;

   static Persistent* add_type(Chi_String, Persistent*);
   static Persistent* clone(Chi_String);
private:
   static Chi_Map<Chi_String, Persistent*> _typemap;
};


#line 144 "PSTREAM.cpp"


template<class X>
void archive(X*& p, PersistentStream& ps)
/* PURPOSE:    archive a persistent pointer on a stream
   RECEIVES:   p - a pointer to an object of a type derived from Persistent
               ps - a persistent stream
*/
{  ps.archive((Persistent*&)p);
}

/*.........................................................................*/

template<class X>
void archive(Chi_Array<X>& a, PersistentStream& ps)
/* PURPOSE:    archive an array of archivable items
   RECEIVES:   a - an array of object of a type that can be archived
               ps - a persistent stream
*/
{  if (ps.is_saving())
      ps << a.low() << " " << a.high() << endl;
   else
   {  int l, h;
      ps >> l >> h;
      a.empty();
      a.grow(l, h);
   }
   for (int i = a.low(); i <= a.high(); i++)
   {  archive( a[i], ps);
   }
}

/*.........................................................................*/

template<class X>
void archive(Chi_PtrArray<X>& a, PersistentStream& ps)
/* PURPOSE:    archive an array of archivable items
   RECEIVES:   a - an array of object of a type that can be archived
               ps - a persistent stream
*/
{  if (ps.is_saving())
      ps << a.low() << " " << a.high() << endl;
   else
   {  int l, h;
      ps >> l >> h;
      a.empty();
      a.grow(l, h);
   }
   for (int i = a.low(); i <= a.high(); i++)
   {  archive( a[i], ps);
   }
}

/*.........................................................................*/

template<class X>
void archive(Chi_NumArray<X>& a, PersistentStream& ps)
/* PURPOSE:    archive an array of archivable items
   RECEIVES:   a - an array of object of a type that can be archived
               ps - a persistent stream
*/
{  if (ps.is_saving())
      ps << a.low() << " " << a.high() << endl;
   else
   {  int l, h;
      ps >> l >> h;
      a.empty();
      a.grow(l, h);
   }
   for (int i = a.low(); i <= a.high(); i++)
   {  archive( a[i], ps);
   }
}

/*.........................................................................*/

#define DEFINE_PERSISTENT(X) \
   Chi_String name() const { return #X; } \
   Persistent* clone() const { return new X(*this); } \
   static Persistent* _pobj;

/*.........................................................................*/

#define IMPLEMENT_PERSISTENT(X) \
   Persistent* X::_pobj = Persistent::add_type(#X, new X);


#endif
