/****************************************************************************
** COPYRIGHT (C):    1994 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          Mastering Object-Oriented Design
** FILE:             shapes2.cpp
** PURPOSE:          Static and dynamic binding
** VERSION           1.0
** LANGUAGE:         Borland C++ 3.1
** TARGET:           IBM PC
** PROGRAMMERS:      Cay S. Horstmann (CSH)
** START DATE:       September 14, 1994
** UPDATE HISTORY:
****************************************************************************/

/* NOTES ********************************************************************

Here we investigate the difference between static and dynamic binding.
The required shape classes and operations are copied from chishape.
Place a // comment at the beginning of the line "#define virtual" to
activate virtual functions.

****************************************************************************/

/* IMPORT ******************************************************************/

#include "chiarray.h"
#include "chigc.h"

/* Place a // comment before the next line to activate virtual functions */
#define virtual

/* TYPES *******************************************************************/

EXPORT class Chi_Shape
/* PURPOSE:    A shape that can be plotted and moved around
*/
{
public:
   virtual void plot(Chi_GraphicsContext& gc) const {};
};

/*-------------------------------------------------------------------------*/

EXPORT class Chi_Point : public Chi_Shape
/* PURPOSE:    A single point
*/
{
public:
   Chi_Point(double xx =0, double yy =0);
   virtual void plot(Chi_GraphicsContext& gc) const;

   double x() const;
   double y() const;

private:
   double _x;
   double _y;
};

/*-------------------------------------------------------------------------*/

EXPORT class Chi_Rectangle : public Chi_Shape
/* PURPOSE:    A rectangle (with sides parallel to the coordinate axes)
*/
{
public:
   Chi_Rectangle();
   Chi_Rectangle(Chi_Point, Chi_Point);

   virtual void plot(Chi_GraphicsContext& gc) const;

private:
   Chi_Point _corner1;
   Chi_Point _corner2;
};

/*-------------------------------------------------------------------------*/

class Chi_FilledRect : public Chi_Rectangle
/* PURPOSE:    A rectangle filled with a pattern
*/
{
public:
   Chi_FilledRect();
   Chi_FilledRect(Chi_Point, Chi_Point, Chi_GraphicsContext::BrushStyle);
   virtual void plot(Chi_GraphicsContext& gc) const;
private:
   Chi_GraphicsContext::BrushStyle _pattern;
};

/* FUNCTIONS ***************************************************************/

int chi_main(int, char**)
{  Chi_GraphicsContext gc;
   Chi_PtrArray<Chi_Rectangle> chart; // an array of pointers
   chart.grow(1,4);

   chart[1] = new Chi_FilledRect(Chi_Point(1.0, 7.0), Chi_Point(2.0, 2.0),
      Chi_GraphicsContext::FDIAG_BRUSH);
   chart[2] = new Chi_Rectangle(Chi_Point(2.0, 7.0), Chi_Point(3.0, 1.0));
   chart[3] = new Chi_FilledRect(Chi_Point(3.0, 7.0),
      Chi_Point(4.0, 3.0), Chi_GraphicsContext::HORIZ_BRUSH);
   chart[4] = new Chi_Rectangle(Chi_Point(0.5, 0.5), Chi_Point(4.5, 7.5));

   for (int i = chart.low(); i <= chart.high(); i++)
      chart[i]->plot(gc);

   gc.close();
   return 0;
}

/*-------------------------------------------------------------------------*/

Chi_Point::Chi_Point(double xx, double yy)
/* RECEIVES:   xx, yy - coordinates
*/
   : _x(xx), _y(yy)
{}

//............................................................................

void Chi_Point::plot(Chi_GraphicsContext& gc) const
/* PURPOSE:    plot this shape
   RECEIVES:   gc - graphics context
*/
{  gc.point(_x, _y);
}

//............................................................................

inline double Chi_Point::x() const
/* RETURNS:    the x-coordinate
*/
{  return _x;
}

//............................................................................

inline double Chi_Point::y() const
/* RETURNS:    the y-coordinate
*/
{  return _y;
}

/*-------------------------------------------------------------------------*/

Chi_Rectangle::Chi_Rectangle() {}

//............................................................................

Chi_Rectangle::Chi_Rectangle(Chi_Point pp, Chi_Point qq)
/* RECEIVES:   pp, qq - two corner points
*/
   : _corner1(pp), _corner2(qq)
{}

//............................................................................

void Chi_Rectangle::plot(Chi_GraphicsContext& gc) const
/* PURPOSE:    plot this shape
   RECEIVES:   gc - graphics context
*/
{  gc.rectangle(_corner1.x(), _corner1.y(), _corner2.x(), _corner2.y());
}

/*-------------------------------------------------------------------------*/

Chi_FilledRect::Chi_FilledRect()
: _pattern(Chi_GraphicsContext::SOLID_BRUSH)
{}

//............................................................................

Chi_FilledRect::Chi_FilledRect(Chi_Point p, Chi_Point q, Chi_GraphicsContext::BrushStyle b)
/* RECEIVES:   p, q - corner points
               b - brush style to use for filling
*/
: _pattern(b), Chi_Rectangle(p, q)
{}

//............................................................................

void Chi_FilledRect::plot(Chi_GraphicsContext& gc) const
/* PURPOSE:    plot this shape
   RECEIVES:   gc - graphics context
*/
{  gc.set_brush_style(_pattern);
   Chi_Rectangle::plot(gc);
   gc.set_brush_style(Chi_GraphicsContext::EMPTY_BRUSH);
}

/***************************************************************************/
