/****************************************************************************
** COPYRIGHT (C):    1994 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          Mastering Object-Oriented Design
** FILE:             shapes3.cpp
** PURPOSE:          Type fields
** VERSION           1.0
** LANGUAGE:         Borland C++ 3.1
** TARGET:           IBM PC
** PROGRAMMERS:      Cay S. Horstmann (CSH)
** START DATE:       September 14, 1994
** UPDATE HISTORY:
****************************************************************************/

/* IMPORT ******************************************************************/

#include "chiarray.h"
#include "chigc.h"

/* TYPES *******************************************************************/

EXPORT class T_Shape
/* PURPOSE:    A shape that can be plotted and moved around
*/
{
public:
   void set_type(char); // called in constructors of derived classes
   char type() const;
   void plot(Chi_GraphicsContext& gc) const {};
   char _type;
};

/*-------------------------------------------------------------------------*/

EXPORT class T_Point : public T_Shape
/* PURPOSE:    A single point
*/
{
public:
   T_Point(double xx =0, double yy =0);
   void plot(Chi_GraphicsContext& gc) const;

   double x() const;
   double y() const;

private:
   double _x;
   double _y;
};

/*-------------------------------------------------------------------------*/

EXPORT class T_Rectangle : public T_Shape
/* PURPOSE:    A rectangle (with sides parallel to the coordinate axes)
*/
{
public:
   T_Rectangle();
   T_Rectangle(T_Point, T_Point);

   void plot(Chi_GraphicsContext& gc) const;

private:
   T_Point _corner1;
   T_Point _corner2;
};

/*-------------------------------------------------------------------------*/

class T_FilledRect : public T_Rectangle
/* PURPOSE:    A rectangle filled with a pattern
*/
{
public:
   T_FilledRect();
   T_FilledRect(T_Point, T_Point, Chi_GraphicsContext::BrushStyle);
   void plot(Chi_GraphicsContext& gc) const;
private:
   Chi_GraphicsContext::BrushStyle _pattern;
};

/* FUNCTIONS ***************************************************************/

int chi_main(int, char**)
{  Chi_GraphicsContext gc;
   Chi_PtrArray<T_Rectangle> chart; // an array of pointers
   chart.grow(1,4);

   chart[1] = new T_FilledRect(T_Point(1.0, 7.0), T_Point(2.0, 2.0),
      Chi_GraphicsContext::FDIAG_BRUSH);
   chart[2] = new T_Rectangle(T_Point(2.0, 7.0), T_Point(3.0, 1.0));
   chart[3] = new T_FilledRect(T_Point(3.0, 7.0),
      T_Point(4.0, 3.0), Chi_GraphicsContext::HORIZ_BRUSH);
   chart[4] = new T_Rectangle(T_Point(0.5, 0.5), T_Point(4.5, 7.5));

   for (int i = chart.low(); i <= chart.high(); i++)
   {  T_Rectangle* r = chart[i];
      switch (r->type())
      {  case 'R':
            r->plot(gc);
            break;
         case 'F':
            ((T_FilledRect*)r)->plot(gc);
            break;
      }
   }
   gc.close();
   return 0;
}

/*-------------------------------------------------------------------------*/

void T_Shape::set_type(char c)
/* PURPOSE:    set the type field
   RECEIVES:   c - the type indicator
   REMARKS:    this function should only be called in derived class
               constructors
*/
{  _type = c;
}

//............................................................................

char T_Shape::type() const
/* RETURNS:    the type indicator
*/
{  return _type;
}

/*-------------------------------------------------------------------------*/

T_Point::T_Point(double xx, double yy)
/* RECEIVES:   xx, yy - coordinates
*/
   : _x(xx), _y(yy)
{  set_type('P');
}

//............................................................................

void T_Point::plot(Chi_GraphicsContext& gc) const
/* PURPOSE:    plot this shape
   RECEIVES:   gc - graphics context
*/
{  gc.point(_x, _y);
}

//............................................................................

inline double T_Point::x() const
/* RETURNS:    the x-coordinate
*/
{  return _x;
}

//............................................................................

inline double T_Point::y() const
/* RETURNS:    the y-coordinate
*/
{  return _y;
}

/*-------------------------------------------------------------------------*/

T_Rectangle::T_Rectangle()
{  set_type('R');
}

//............................................................................

T_Rectangle::T_Rectangle(T_Point pp, T_Point qq)
/* RECEIVES:   pp, qq - two corner points
*/
   : _corner1(pp), _corner2(qq)
{  set_type('R');
}

//............................................................................

void T_Rectangle::plot(Chi_GraphicsContext& gc) const
/* PURPOSE:    plot this shape
   RECEIVES:   gc - graphics context
*/
{  gc.rectangle(_corner1.x(), _corner1.y(), _corner2.x(), _corner2.y());
}

/*-------------------------------------------------------------------------*/

T_FilledRect::T_FilledRect()
: _pattern(Chi_GraphicsContext::SOLID_BRUSH)
{  set_type('F');
}

//............................................................................

T_FilledRect::T_FilledRect(T_Point p, T_Point q, Chi_GraphicsContext::BrushStyle b)
/* RECEIVES:   p, q - corner points
               b - brush style to use for filling
*/
: _pattern(b), T_Rectangle(p, q)
{  set_type('F');
}

//............................................................................

void T_FilledRect::plot(Chi_GraphicsContext& gc) const
/* PURPOSE:    plot this shape
   RECEIVES:   gc - graphics context
*/
{  gc.set_brush_style(_pattern);
   T_Rectangle::plot(gc);
   gc.set_brush_style(Chi_GraphicsContext::EMPTY_BRUSH);
}

/***************************************************************************/
