/****************************************************************************
** COPYRIGHT (C):    1994 Cay S. Horstmann. All Rights Reserved.
** PROJECT:          Mastering Object-Oriented Design
** FILE:             simfrmwk.cpp
** PURPOSE:          A simulation framework
** VERSION           1.0
** LANGUAGE:         Borland C++ 3.0
** TARGET:           IBM PC
** PROGRAMMER:       Cay Horstmann (CSH)
** START DATE:       9-18-94 (CSH)
** UPDATE HISTORY:
****************************************************************************/

/* IMPORT ******************************************************************/

#ifndef EXPORT
#include "chisetup.h"
#endif

#include <math.h>
EXPORT
#include "chiprioq.h"
#include "simfrmwk.h"

/* TYPES *******************************************************************/

#ifndef EXPORT

class Simulation;

/*-------------------------------------------------------------------------*/

typedef double Time; // measured in minutes

/*-------------------------------------------------------------------------*/

class Event
/* PURPOSE:    An event that is scheduled for execution at some time
*/
{
public:
   Event(Time t);

   virtual void process(Simulation&) = 0;
   Time time() const;
   virtual ~Event() {}
private:
   Time _time;
};

/*-------------------------------------------------------------------------*/

class Simulation
/* PURPOSE:    A generic event-driven simulation
*/
{
public:
   Simulation();
   void run(Time from, Time to);
   void schedule(Event*);
   virtual void initialize() = 0;
   virtual void update() = 0;
   virtual void terminate() = 0;

   Time now() const;

   virtual ~Simulation() {}

private:
   Time _now;
   Chi_PtrPriorityQueue<Event> _event_queue;
};

#endif

/* GLOBALS *****************************************************************/

/* the following are for computing expdist--see the description of the
   algorithm in Knuth
*/

static int kmax = 0;
const int KMAX = 30;
static double q[KMAX];

const double LN2 = 0.693147180559945309417;

/* FUNCTIONS ***************************************************************/

static void precompute()
/* PURPOSE:    Precompute q array for expdist
*/
{  double s = 1;
   int k = 0;
   q[0] = 0;
   while (q[ k ] <= (1 - 1/double(RAND_MAX + 1)) && k < KMAX)
   {  k++;
      s = s * LN2 / k;
      q[k] = q[k-1] + s;
   }
   kmax = k;
}

/*.........................................................................*/

EXPORT double expdist(double mean)
/* PURPOSE:    Compute exponentially distributed random numbers
   RECEIVES:   mean - the mean of the number sequence
   RETURNS:    a random number
   REMARKS:    This algorithm, due to Marsaglia, Sibuya and Ahrens, is from
               Knuth, Seminumerical Algorithms, Addison-Wesley 1981, p. 128
               Note that Knuth has two typos: j should start at 0, and in
               step 4 set X to mean (j ln 2 + V), not mean (j + V)ln 2.
*/
{  if (kmax == 0) precompute();
   int iu = rand();
   const unsigned mask = ((unsigned int)RAND_MAX + 1)>>1;
      // assumption: this is a power of 2!!!
   int j = 0;
   while (mask & iu)
   {  iu <<= 1;
      j++;
   }
   iu <<= 1;
   double u = iu/double((unsigned int)RAND_MAX+1);
   if (u < LN2) return mean * (j *LN2 + u);

   int k = 2;
   while (u >= q[k] && k < kmax) k++;
   unsigned uv = (unsigned int)RAND_MAX+1;
   for (int i = 0; i < k; i++)
   {  unsigned uui = rand();
      if (uui < uv) uv = uui;
   }
   double v = uv/double((unsigned int)RAND_MAX+1);
   return mean * (j * LN2 + v);
}

/*.........................................................................*/

static int event_compare(const Event* a, const Event* b)
/* PURPOSE:    compare two events by their time stamp
   RECEIVES:   a, b - pointers to two events
   RETURNS:    -1 if a is scheduled earlier than b, 0 if they are
               scheduled at the same time, 1 otherwise
*/
{  Time d = a->time() - b->time();
   return (d > 0) - (d < 0);
}

/*-------------------------------------------------------------------------*/

Event::Event(Time t)
/* RECEIVES:   t - the execution time of this event
*/
: _time(t)
{}

/*.........................................................................*/

Time Event::time() const
/* RETURNS:    the time at which the event is to be executed
*/
{  return _time;
}

/*-------------------------------------------------------------------------*/

Simulation::Simulation()
: _event_queue(event_compare)
{}

/*.........................................................................*/

void Simulation::run(Time from, Time to)
/* PURPOSE:    run the simulation
   RECEIVES:   from, to - the start and end times
*/
{  _now = from;
   initialize();
   while (_event_queue.length() > 0 && _now <= to)
   {  Event* event = _event_queue.remove();
      _now = event->time();
      event->process(self);
      delete event;
      update();
   }
   terminate();
}

/*.........................................................................*/

void Simulation::schedule(Event* e)
/* PURPOSE:    Queue an event to schedule it for execution
   RECEIVES:   e - the event
*/
{  _event_queue.insert(e);
}

/*.........................................................................*/

#ifndef EXPORT

inline Time Simulation::now() const
/* RETURNS:    the current simulation time
*/
{  return _now;
}

#endif

/***************************************************************************/
